from typing import TYPE_CHECKING, Any, Callable

from functools import wraps
from inspect import getattr_static

if TYPE_CHECKING:
    from machinable.element import Element


class Mixin:
    """Mixin base class"""


class bind:
    """
    Allows to dynamically extend object instances

    # Example
    ```python
    class Extension:
        def greet(self):
            # write an extension for Example class
            # note that self refers to the instance we are extending
            print(self.hello)

    class Example:
        def __init__(self):
            self.hello = 'hello world'
            # extend dynamically
            self.extension = Mixin(self, Extension, 'extension')

    Example().extension.greet()
    >>> 'hello world'
    ```
    """

    def __init__(self, target: Any, mixin_class: Any, name: str):
        self._binding_mixin_target = target
        self._binding_mixin_class = mixin_class
        self._binding_mixin_name = name

    def __getattr__(self, item):
        # forward dynamically into mix-in class
        attribute = getattr(self._binding_mixin_class, item, None)

        if attribute is None:
            raise AttributeError(
                f"'{self._binding_mixin_class.__name__}' has no attribute '{item}'"
            )

        if isinstance(attribute, property):
            return attribute.fget(self._binding_mixin_target)

        if not callable(attribute):
            return attribute

        if isinstance(
            getattr_static(self._binding_mixin_class, item), staticmethod
        ):
            return attribute

        # if attribute is non-static method we decorate it to pass in the controller

        def bound_method(*args, **kwargs):
            return attribute(self._binding_mixin_target, *args, **kwargs)

        return bound_method


def mixin(f: Callable) -> Any:
    @property
    @wraps(f)
    def _wrapper(self: "Element"):
        name = f.__name__
        if name not in self.__mixins__:
            mixin_class = f(self)
            if isinstance(mixin_class, str):
                from machinable.project import Project, import_element

                mixin_class = import_element(
                    Project.get().path(), mixin_class, Mixin
                )

            self.__mixins__[name] = bind(self, mixin_class, name)

        # assign to __mixin__ for reference
        self.__mixin__ = self.__mixins__[name]

        return self.__mixins__[name]

    return _wrapper



from typing import TYPE_CHECKING, List, Union

import os

from machinable import schema
from machinable.element import get_lineage
from machinable.index import Index
from machinable.interface import Interface
from machinable.types import VersionType
from machinable.utils import load_file


class Storage(Interface):
    kind = "Storage"
    default = None

    def __init__(
        self,
        version: VersionType = None,
    ):
        super().__init__(version=version)
        self.__model__ = schema.Storage(
            kind=self.kind,
            module=self.__model__.module,
            config=self.__model__.config,
            version=self.__model__.version,
            lineage=get_lineage(self),
        )

    def upload(
        self, interface: "Interface", related: Union[bool, int] = True
    ) -> None:
        """
        Upload the interface to the storage

        interface: Interface
        related: bool | int
            - 0/False: Do not save related interfaces
            - 1/True: Save immediately related interfaces
            - 2: Save related interfaces and their related interfaces
        """
        if self.contains(interface.uuid):
            self.update(interface)
        else:
            self.commit(interface)

        if related:
            for r in interface.related(deep=int(related) == 2).all():
                self.upload(r, related=False)

    def download(
        self,
        uuid: str,
        destination: Union[str, None] = None,
        related: Union[bool, int] = True,
    ) -> List[str]:
        """
        Download to destination

        uuid: str
            Primary interface UUID
        destination: str | None
            If None, download will be imported into active index. If directory filepath, download will be placed
            in directory without import to index
        related: bool | int
            - 0/False: Do not upload related interfaces
            - 1/True: Upload immediately related interfaces
            - 2: Upload related interfaces and their related interfaces

        Returns:
            List of downloaded directories
        """
        index = None
        download_directory = destination
        if destination is None:
            index = Index.get()
            download_directory = index.config.directory

        retrieved = []

        # retrieve primary
        local_directory = os.path.join(download_directory, uuid)
        if self.retrieve(uuid, local_directory):
            retrieved.append(local_directory)

        # retrieve related
        if related:
            related_uuids = set()
            for r in load_file(
                [local_directory, "related", "metadata.jsonl"], []
            ):
                related_uuids.add(r["uuid"])
                related_uuids.add(r["related_uuid"])

            for r in related_uuids:
                if r == uuid:
                    continue
                retrieved.extend(
                    self.download(
                        r,
                        destination,
                        related=2 if int(related) == 2 else False,
                    )
                )

        # import to index
        if index is not None:
            for directory in retrieved:
                index.import_directory(directory, relations=bool(related))

        return retrieved

    def commit(self, interface: "Interface") -> None:
        ...

    def update(self, interface: "Interface") -> None:
        return self.commit(interface)

    def contains(self, uuid: str) -> bool:
        ...
        return False

    def retrieve(self, uuid: str, local_directory: str) -> bool:
        ...
        return False

    def search_for(self, interface: "Interface") -> List[str]:
        raise NotImplementedError()


def fetch(uuid: str, directory: str) -> bool:
    available = False
    for storage in Storage.connected():
        if storage.retrieve(uuid, directory):
            available = True
            break

    return available



from typing import Dict

from machinable.config import to_dict
from machinable.interface import Interface


class Scope(Interface):
    kind = "Scope"
    default = None

    def __call__(self) -> Dict:
        return to_dict(self.config._update_)



class MachinableError(Exception):
    """All of machinable exception inherit from this baseclass"""


class ConfigurationError(MachinableError):
    """Invalid configuration

    Bases: MachinableError"""


class DependencyMissing(MachinableError, ImportError):
    """Missing optional dependency

    Bases: ImportError"""


class StorageError(MachinableError):
    """Storage error

    Bases: MachinableError"""


class ExecutionFailed(MachinableError):
    """Execution failed

    Bases: MachinableError"""


class ComponentException(MachinableError):
    """Component exception

    Bases: MachinableError"""



# This file contains modified 3rd party source code from
# https://github.com/sdispater/backpack/blob/master/backpack/collections/base_collection.py.
# The copyright and license agreement can be found in the ThirdPartyNotices.txt file at the root of this repository.

from typing import TYPE_CHECKING, Any, Callable, Optional, Union

import copy
from functools import reduce
from json import dumps
from pprint import pprint

from machinable.types import VersionType

long = int
unicode = str
basestring = str


if TYPE_CHECKING:
    from machinable.component import Component


def _get_value(val):
    if callable(val):
        return val()

    return val


def data_get(target, key, default=None):
    """
    Get an item from a list, a dict or an object using "dot" notation.

    :param target: The target element
    :type target: list or dict or object

    :param key: The key to get
    :type key: string or list

    :param default: The default value
    :type default: mixed

    :rtype: mixed
    """
    if key is None:
        return target

    if not isinstance(key, list):
        key = key.split(".")

    for segment in key:
        if isinstance(target, (list, tuple)):
            try:
                target = target[segment]
            except IndexError:
                return _get_value(default)
        elif isinstance(target, dict):
            try:
                target = target[segment]
            except IndexError:
                return _get_value(default)
        else:
            try:
                target = target[segment]
            except (IndexError, KeyError, TypeError):
                try:
                    target = getattr(target, segment)
                except AttributeError:
                    return _get_value(default)

    return target


class Collection:
    def __init__(self, items=None):
        """Creates a new Collection

        # Arguments
        items: ``list``|``Collection``|``map`` of items to collect
        """
        if items is None:
            items = []
        else:
            items = self._get_items(items)

        self._items = items

    @property
    def items(self):
        """Items of the collection"""
        return self._items

    @classmethod
    def make(cls, items=None):
        """Create a new Collection instance if the value isn't one already

        # Arguments
        items: ``list``|``Collection``|``map`` of items to collect
        """
        if isinstance(items, cls):
            return items

        return cls(items)

    def all(self):
        """Get all of the items in the collection.

        Returns the underlying list represented by the
        collection

        ``` python
        Collection([1, 2, 3]).all()

        # [1, 2, 3]
        ```
        """
        return self.items

    def avg(self, key=None):
        """Get the average value of a given key.

        # Arguments
        key: The key to get the average for

        ``` python
        Collection([1, 2, 3, 4, 5]).avg()
        # 3
        ```

        If the collection contains nested objects or dictionaries, you must pass
        a key to use for determining which values to calculate the average:

        ``` python
        collection = Collection([
            {'name': 'JavaScript: The Good Parts', 'pages': 176},
            {'name': 'JavaScript: The Defnitive Guide', 'pages': 1096}
        ])
        # 636
        collection.avg('pages')
        ```
        """
        count = self.count()

        if count:
            return self.sum(key) / count

    def chunk(self, size):
        """Chunk the underlying collection.

        The `chunk` method breaks the collection into multiple, smaller
        collections of a given size:

        ``` python
        collection = Collection([1, 2, 3, 4, 5, 6, 7])

        chunks = collection.chunk(4)

        chunks.serialize()

        # [[1, 2, 3, 4], [5, 6, 7]]
        ```

        # Arguments
        size: The chunk size
        """
        chunks = self._chunk(size)

        return self.__class__(list(map(self.__class__, chunks)))

    def _chunk(self, size):
        items = self.items
        return [items[i : i + size] for i in range(0, len(items), size)]

    def count(self):
        """Returns the total number of items in the collection:

        ``` python
        collection = Collection([1, 2, 3, 4])

        collection.count()

        # 4
        ```

        The `len` function can also be used:

        ``` python
        len(collection)

        # 4
        ```
        """
        return len(self._items)

    def contains(self, key, value=None):
        """Determines if an element is in the collection

        # Arguments
        key: ``Integer``|``String``|``callable`` The element
        value: The value of the element

        ``` python
        collection = Collection(['foo', 'bar'])

        collection.contains('foo')

        # True
        ```

        You can also use the `in` keyword:

        ``` python
        'foo' in collection

        # True
        ```

        You can also pass a key / value pair to the `contains` method, which
        will determine if the given pair exists in the collection:

        ``` python
        collection = Collection([
            {'name': 'John', 'id': 1},
            {'name': 'Jane', 'id': 2}
        ])

        collection.contains('name', 'Simon')

        # False
        ```

        Finally, you may also pass a callback to the `contains` method to
        perform your own truth test:

        ``` python
        collection = Collection([1, 2, 3, 4, 5])

        collection.contains(lambda item: item > 5)

        # False
        ```
        """
        if value is not None:
            return self.contains(lambda x: data_get(x, key) == value)

        if self._use_as_callable(key):
            return self.first(key) is not None

        return key in self.items

    def __contains__(self, item):
        return self.contains(item)

    def collapse(self):
        """Collapses a collection of lists into a flat collection

        ``` python
        collection = Collection([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        collapsed = collection.collapse()
        collapsed.all()
        # [1, 2, 3, 4, 5, 6, 7, 8, 9]
        ```
        """
        results = []

        items = self.items

        for values in items:
            if isinstance(values, Collection):
                values = values.all()

            results += values

        return self.__class__(results)

    def diff(self, items):
        """Compares the collection against another collection, a `list` or a `dict`

        # Arguments
        items: The items to diff with

        ``` python
        collection = Collection([1, 2, 3, 4, 5])
        diff = collection.diff([2, 4, 6, 8])
        diff.all()
        # [1, 3, 5]
        ```
        """
        return self.__class__([i for i in self.items if i not in items])

    def each(self, callback):
        """Iterates over the items in the collection and passes each item to a given callback

        # Arguments
        callback: ``callable`` The callback to execute

        ``` python
        collection = Collection([1, 2, 3])
        collection.each(lambda x: x + 3)
        ```

        Return `False` from your callback to break out of the loop:

        ``` python
        observations.each(lambda data: data.save() if data.name == 'mnist' else False)
        ```

        ::: tip
        It only applies the callback but does not modify the collection's items.
        Use the [transform()](#transform) method to
        modify the collection.
        :::
        """
        items = self.items

        for item in items:
            if callback(item) is False:
                break

        return self

    def every(self, step, offset=0):
        """Create a new collection consisting of every n-th element.

        # Arguments
        step: ``int`` The step size
        offset: ``int`` The start offset

        ``` python
        collection = Collection(['a', 'b', 'c', 'd', 'e', 'f'])

        collection.every(4).all()

        # ['a', 'e']
        ```

        You can optionally pass the offset as the second argument:

        ``` python
        collection.every(4, 1).all()

        # ['b', 'f']
        ```
        """
        new = []

        for position, item in enumerate(self.items):
            if position % step == offset:
                new.append(item)

        return self.__class__(new)

    def without(self, *keys):
        """Get all items except for those with the specified keys.

        # Arguments
        keys: ``tuple`` The keys to remove
        """
        items = copy.copy(self.items)

        keys = reversed(sorted(keys))

        for key in keys:
            del items[key]

        return self.__class__(items)

    def only(self, *keys):
        """
        Get the items with the specified keys.

        # Arguments
        keys: ``tuple`` The keys to keep
        """
        items = []

        for key, value in enumerate(self.items):
            if key in keys:
                items.append(value)

        return self.__class__(items)

    def filter(self, callback=None):
        """Filters the collection by a given callback, keeping only those items that pass a given truth test

        # Arguments
        callback: ``callable``|``None`` The filter callback

        ``` python
        collection = Collection([1, 2, 3, 4])

        filtered = collection.filter(lambda item: item > 2)

        filtered.all()

        # [3, 4]
        ```
        """
        if callback:
            return self.__class__(list(filter(callback, self.items)))

        return self.__class__(list(filter(None, self.items)))

    def where(self, key, value):
        """Filter items by the given key value pair.

        # Arguments
        key: The key to filter by
        value: The value to filter by

        ``` python
        collection = Collection([
            {'name': 'Desk', 'price': 200},
            {'name': 'Chair', 'price': 100},
            {'name': 'Bookcase', 'price': 150},
            {'name': 'Door', 'price': 100},
        ])

        filtered = collection.where('price', 100)

        filtered.all()

        # [
        #     {'name': 'Chair', 'price': 100},
        #     {'name': 'Door', 'price': 100}
        # ]
        ```
        """
        return self.filter(lambda item: data_get(item, key) == value)

    def first(self, callback=None, default=None):
        """Returns the first element in the collection that passes a given truth test

        # Arguments
        callback: Optional callable truth condition to find first element
        default: A default value

        ``` python
        collection = Collection([1, 2, 3, 4])

        collection.first(lambda item: item > 2)

        # 3
        ```

        You can also call the `first` method with no arguments to get the first
        element in the collection. If the collection is empty, `None` is
        returned:

        ``` python
        collection.first()

        # 1
        ```
        """
        if callback is not None:
            for val in self.items:
                if callback(val):
                    return val

            return _get_value(default)

        if len(self.items) > 0:
            return self.items[0]
        else:
            return default

    def flatten(self):
        """Flattens a multi-dimensional collection into a single dimension

        ``` python
        collection = Collection([1, 2, [3, 4, 5, {'foo': 'bar'}]])

        flattened = collection.flatten()

        flattened.all()

        # [1, 2, 3, 4, 5, 'bar']
        ```
        """

        def _flatten(d):
            if isinstance(d, dict):
                for v in d.values():
                    yield from _flatten(v)
            elif isinstance(d, list):
                for list_v in d:
                    yield from _flatten(list_v)
            else:
                yield d

        return self.__class__(list(_flatten(self.items)))

    def forget(self, *keys):
        """Remove an item from the collection by key.

        # Arguments
        keys: The keys to remove

        ``` python
        collection = Collection([1, 2, 3, 4, 5])
        collection.forget(1)
        collection.all()
        # [1, 3, 4, 5]
        ```

        ::: warning
        Unlike most other collection methods, `forget` does not return a new
        modified collection; it modifies the collection it is called on.
        :::
        """
        keys = reversed(sorted(keys))

        for key in keys:
            del self[key]

        return self

    def get(self, key, default=None):
        """Returns the item at a given key. If the key does not exist, `None` is returned

        # Arguments
        key: The index of the element
        default: The default value to return

        ``` python
        collection = Collection([1, 2, 3])
        collection.get(3)
        # None
        ```

        You can optionally pass a default value as the second argument:

        ``` python
        collection = Collection([1, 2, 3])
        collection.get(3, 'default-value')
        # default-value
        ```
        """
        try:
            return self.items[key]
        except IndexError:
            return _get_value(default)

    def implode(self, value, glue=""):
        """Joins the items in a collection. Its arguments depend on the type of items in the collection.

        # Arguments
        value: The value
        glue: The glue

        If the collection contains dictionaries or objects, you must pass the
        key of the attributes you wish to join, and the "glue" string you wish
        to place between the values:

        ``` python
        collection = Collection([
            {'account_id': 1, 'product': 'Desk'},
            {'account_id': 2, 'product': 'Chair'}
        ])

        collection.implode('product', ', ')

        # Desk, Chair
        ```

        If the collection contains simple strings, simply pass the "glue" as the
        only argument to the method:

        ``` python
        collection = Collection(['foo', 'bar', 'baz'])

        collection.implode('-')

        # foo-bar-baz
        ```
        """
        first = self.first()

        if not isinstance(first, basestring):
            return glue.join(self.pluck(value).all())

        return value.join(self.items)

    def last(self, callback=None, default=None):
        """Returns the last element in the collection that passes a given truth test

        # Arguments
        callback: Optional ``callable`` truth condition
        default: The default value

        ``` python
        collection = Collection([1, 2, 3, 4])

        last = collection.last(lambda item: item < 3)

        # 2
        ```

        You can also call the `last` method with no arguments to get the last
        element in the collection. If the collection is empty, `None` is
        returned:

        ``` python
        collection.last()

        # 4
        ```
        """
        if callback is not None:
            for val in reversed(self.items):
                if callback(val):
                    return val

            return _get_value(default)

        if len(self.items) > 0:
            return self.items[-1]
        else:
            return default

    def pluck(self, value, key=None):
        """Retrieves all of the collection values for a given key

        # Arguments
        value: Value
        key: Optional key

        ``` python
        collection = Collection([
            {'product_id': 1, 'product': 'Desk'},
            {'product_id': 2, 'product': 'Chair'}
        ])

        plucked = collection.pluck('product')

        plucked.all()

        # ['Desk', 'Chair']
        ```

        You can also specify how you wish the resulting collection to be keyed:

        ``` python
        plucked = collection.pluck('name', 'product_id')

        plucked

        # {1: 'Desk', 2: 'Chair'}
        ```
        """
        if self.items is None:
            return self.__class__([])

        if key:
            return dict(
                map(
                    lambda x: (data_get(x, key), data_get(x, value)), self.items
                )
            )
        else:
            results = list(map(lambda x: data_get(x, value), self.items))

        return self.__class__(results)

    def map(self, callback):
        """Iterates through the collection and passes each value to the given callback.
        The callback is free to modify the item and return it, thus forming a new collection of modified items

        # Arguments
        callback: The map function

        ``` python
        collection = Collection([1, 2, 3, 4])

        multiplied = collection.map(lambda item: item * 2)

        multiplied.all()

        # [2, 4, 6, 8]
        ```

        ::: warning
        Like most other collection methods, `map` returns a new `Collection`
        instance; it does not modify the collection it is called on. If you want
        to transform the original collection, use the [transform](#transform)
        method.
        :::
        """
        return self.__class__(list(map(callback, self.items)))

    def max(self, key=None):
        """Get the max value of a given key.

        # Arguments
        key: The key
        """

        def _max(result, item):
            val = data_get(item, key)

            if result is None or val > result:
                return val

            return result

        return self.reduce(_max)

    def min(self, key=None):
        """Get the min value of a given key.

        key: The key
        """

        def _min(result, item):
            val = data_get(item, key)

            if result is None or val < result:
                return val

            return result

        return self.reduce(_min)

    def pop(self, key=None):
        """Removes and returns the last item from the collection.
        If no index is specified, returns the last item.

        # Arguments
        key: The index of the item to return


        ``` python
        collection = Collection([1, 2, 3, 4, 5])
        collection.pop()
        # 5

        collection.all()
        # [1, 2, 3, 4]
        ```
        """
        if key is None:
            key = -1

        return self.items.pop(key)

    def prepend(self, value):
        """Adds an item to the beginning of the collection

        # Arguments
        value: The value to push

        ``` python
        collection = Collection([1, 2, 3, 4])

        collection.prepend(0)

        collection.all()

        # [0, 1, 2, 3, 4]
        ```
        """
        self.items.insert(0, value)

        return self

    def append(self, value):
        """Add an item onto the end of the collection.

        # Arguments
        value: The value to push

        ``` python
        collection = Collection([1, 2, 3, 4])

        collection.push(5)

        collection.all()

        # [1, 2, 3, 4, 5]
        ```
        """
        self.items.append(value)

        return self

    def pull(self, key, default=None):
        """Removes and returns an item from the collection by its key

        # Arugments
        key: The key
        default: The default value

        ``` python
        collection = Collection([1, 2, 3, 4])

        collection.pull(1)

        collection.all()

        # [1, 3, 4]
        ```
        """
        val = self.get(key, default)

        self.forget(key)

        return val

    def put(self, key, value):
        """Sets the given key and value in the collection

        # Arguments
        key: The key
        value: The value

        ``` python
        collection = Collection([1, 2, 3, 4])
        collection.put(1, 5)
        collection.all()

        # [1, 5, 3, 4]
        ```

        ::: tip
        It is equivalent to ``collection[1] = 5``
        :::
        """
        self[key] = value

        return self

    def reduce(self, callback, initial=None):
        """Reduces the collection to a single value, passing the result of each iteration into the subsequent iteration

        # Arguments
        callback: The callback
        initial: The initial value

        ``` python
        collection = Collection([1, 2, 3])

        collection.reduce(lambda result, item: (result or 0) + item)

        # 6
        ```

        The value for `result` on the first iteration is `None`; however, you
        can specify its initial value by passing a second argument to reduce:

        ``` python
        collection.reduce(lambda result, item: result + item, 4)

        # 10
        ```
        """
        return reduce(callback, self.items, initial)

    def reject(self, callback):
        """Filters the collection using the given callback. The callback should return `True` for any items it wishes
        to remove from the resulting collection

        # Arguments
        callback: The truth test

        ``` python
        collection = Collection([1, 2, 3, 4])

        filtered = collection.reject(lambda item: item > 2)

        filtered.all()

        # [1, 2]
        ```

        For the inverse of `reject`, see the [filter](#filter) method.
        """
        if self._use_as_callable(callback):
            return self.filter(lambda item: not callback(item))

        return self.filter(lambda item: item != callback)

    def reverse(self):
        """Reverses the order of the collection's items

        ``` python
        collection = Collection([1, 2, 3, 4, 5])
        reverse = collection.reverse()
        reverse.all()
        # [5, 4, 3, 2, 1]
        ```
        """
        return self.__class__(list(reversed(self.items)))

    def sort(self, callback=None, reverse=False):
        """Sorts the collection

        # Arguments
        callback: Sort callable
        reverse: True for reversed sort order

        ``` python
        collection = Collection([5, 3, 1, 2, 4])

        sorted = collection.sort()

        sorted.all()

        # [1, 2, 3, 4, 5]
        ```
        """
        items = self.items

        if callback:
            return self.__class__(sorted(items, key=callback, reverse=reverse))
        else:
            return self.__class__(sorted(items, reverse=reverse))

    def sum(self, callback=None):
        """Returns the sum of all items in the collection

        callback: The callback

        ``` python
        Collection([1, 2, 3, 4, 5]).sum()

        # 15
        ```

        If the collection contains dictionaries or objects, you must pass a key
        to use for determining which values to sum:

        ``` python
        collection = Collection([
            {'name': 'JavaScript: The Good Parts', 'pages': 176},
            {'name': 'JavaScript: The Defnitive Guide', 'pages': 1096}
        ])

        collection.sum('pages')

        # 1272
        ```

        In addition, you can pass your own callback to determine which values of
        the collection to sum:

        ``` python
        collection = Collection([
            {'name': 'Chair', 'colors': ['Black']},
            {'name': 'Desk', 'colors': ['Black', 'Mahogany']},
            {'name': 'Bookcase', 'colors': ['Red', 'Beige', 'Brown']}
        ])

        collection.sum(lambda product: len(product['colors']))

        # 6
        ```
        """
        if callback is None:
            return sum(self.items)

        callback = self._value_retriever(callback)

        return self.reduce(lambda result, item: (result or 0) + callback(item))

    def take(self, limit):
        """
        Take the first or last n items.

        # Arguments
        limit: The number of items to take

        ``` python
        collection = Collection([0, 1, 2, 3, 4, 5])
        chunk = collection.take(3)
        chunk.all()
        # [0, 1, 2]
        ```

        You can also pass a negative integer to take the specified amount of
        items from the end of the collection:

        ``` python
        chunk = collection.chunk(-2)
        chunk.all()
        # [4, 5]
        ```
        """
        if limit < 0:
            return self[limit:]

        return self[:limit]

    def unique(self, key=None):
        """Returns all of the unique items in the collection

        # Arguments
        key: The key to check uniqueness on

        ``` python
        collection = Collection([1, 1, 2, 2, 3, 4, 2])

        unique = collection.unique()

        unique.all()

        # [1, 2, 3, 4]
        ```

        When dealing with dictionaries or objects, you can specify the key used
        to determine uniqueness:

        ``` python
        collection = Collection([
            {'name': 'iPhone 6', 'brand': 'Apple', 'type': 'phone'},
            {'name': 'iPhone 5', 'brand': 'Apple', 'type': 'phone'},
            {'name': 'Apple Watch', 'brand': 'Apple', 'type': 'watch'},
            {'name': 'Galaxy S6', 'brand': 'Samsung', 'type': 'phone'},
            {'name': 'Galaxy Gear', 'brand': 'Samsung', 'type': 'watch'}
        ])

        unique = collection.unique('brand')

        unique.all()

        # [
        #     {'name': 'iPhone 6', 'brand': 'Apple', 'type': 'phone'},
        #     {'name': 'Galaxy S6', 'brand': 'Samsung', 'type': 'phone'}
        # ]
        ```

        You can also pass your own callback to determine item uniqueness:

        ``` python
        unique = collection.unique(lambda item: item['brand'] + item['type'])

        unique.all()

        # [
        #     {'name': 'iPhone 6', 'brand': 'Apple', 'type': 'phone'},
        #     {'name': 'Apple Watch', 'brand': 'Apple', 'type': 'watch'},
        #     {'name': 'Galaxy S6', 'brand': 'Samsung', 'type': 'phone'},
        #     {'name': 'Galaxy Gear', 'brand': 'Samsung', 'type': 'watch'}
        # ]
        ```
        """
        if key is None:
            seen = set()
            seen_add = seen.add

            return self.__class__(
                [x for x in self.items if not (x in seen or seen_add(x))]
            )

        key = self._value_retriever(key)

        exists = []

        def _check(item):
            id_ = key(item)
            if id_ in exists:
                return True

            exists.append(id_)

        return self.reject(_check)

    def zip(self, *items):
        """Merges together the values of the given list with the values of the collection at the corresponding index

        # Argument
        *items: Zip items

        ``` python
        collection = Collection(['Chair', 'Desk'])
        zipped = collection.zip([100, 200])
        zipped.all()
        # [('Chair', 100), ('Desk', 200)]
        ```
        """
        return self.__class__(list(zip(self.items, *items)))

    def empty(self):
        """Returns `True` if the collection is empty; otherwise, `False` is returned"""
        return self.count() == 0

    def merge(self, items):
        """Merges the given list into the collection

        # Arguments
        items: The items to merge

        ``` python
        collection = Collection(['Desk', 'Chair'])
        collection.merge(['Bookcase', 'Door'])
        collection.all()
        # ['Desk', 'Chair', 'Bookcase', 'Door']
        ```

        ::: warning
        Unlike most other collection methods, `merge` does not return a new
        modified collection; it modifies the collection it is called on.
        :::
        """
        if isinstance(items, Collection):
            items = items.all()

        if not isinstance(items, list):
            raise ValueError("Unable to merge uncompatible types")

        self._items += items

        return self

    def transform(self, callback):
        """Transform each item in the collection using a callback.

        Iterates over the collection and calls the given callback with each item in the collection.
        The items in the collection will be replaced by the values returned by the callback.

        # Arguments
        callback: The callback

        ``` python
        collection = Collection([1, 2, 3, 4, 5])
        collection.transform(lambda item: item * 2)
        collection.all()

        # [2, 4, 6, 8, 10]
        ```

        ::: warning
        Unlike most other collection methods, `transform` modifies the
        collection itself. If you wish to create a new collection instead, use
        the [map](#map) method.
        :::
        """
        self._items = self.map(callback).all()

        return self

    def serialize(self):
        """Converts the collection into a `list`

        ``` python
        collection = Collection([User.find(1)])
        collection.serialize()
        # [{'id': 1, 'name': 'John'}]
        ```

        ::: warning
        `serialize` also converts all of its nested objects. If you want to get
        the underlying items as is, use the [all](#all) method instead.
        :::
        """

        def _serialize(value):
            if hasattr(value, "serialize"):
                return value.serialize()
            elif hasattr(value, "as_dict"):
                return value.as_dict()
            else:
                return value

        return list(map(_serialize, self.items))

    def pprint(self, pformat="json"):
        if pformat == "json":
            print(dumps(self.all(), indent=4, sort_keys=True, default=str))
        else:
            pprint(self.all())

    def as_json(self, **options):
        """Converts the collection into JSON

        # Arguments
        options: JSON encoding options

        ``` python
        collection = Collection([{'name': 'Desk', 'price': 200}])

        collection.as_json()

        # '[{"name": "Desk", "price": 200}]'
        ```
        """
        return dumps(self.serialize(), **options)

    def as_numpy(self):
        """Converts the collection into a numpy array"""
        import numpy as np

        return np.array(self.items)

    def as_dataframe(self):
        """Returns collection as Pandas dataframe"""
        data = {k: [] for k in self._items[0].serialize().keys()}
        for item in self._items:
            for k, v in item.serialize().items():
                data[k].append(v)
        import pandas

        return pandas.DataFrame.from_dict(data)

    def as_table(self, mode="html", headers=(), **kwargs):
        """Converts the collection into a table

        # Arguments
        mode: String 'html' or any other mode of the tabulate package
        headers: Optional header row
        **kwargs: Options to pass to tabulate
        """
        if len(self) == 0:
            return ""

        try:
            iter(self.first())
        except TypeError:
            raise ValueError(
                "Collection items are not iterable to form columns"
            )

        try:
            from tabulate import tabulate

            return tabulate(
                self.items, headers=headers, tablefmt=mode, **kwargs
            )
        except ImportError:
            if mode != "html":
                raise ValueError(
                    "The tabulate package is required to render non-html tables"
                )

            return "<table><tr>{}</tr></table>".format(
                "</tr><tr>".join(
                    "<td>{}</td>".format("</td><td>".join(str(_) for _ in row))
                    for row in self
                )
            )

    def pluck_or_none(self, value, key=None, none=None):
        """Pluck method that returns None if key is not present

        # Arguments
        value: Value
        key: Key
        none: Return value if key is not present
        """
        try:
            return self.pluck(value, key)
        except KeyError:
            return Collection([none] * len(self))

    def pluck_or_nan(self, value, key=None):
        """Pluck method that returns NaNs if key is not present

        # Arguments
        value: Value
        key: Key
        """
        return self.pluck_or_none(value, key, none=float("nan"))

    def section(self, of, reduce=None):
        """Performs horizontal reduce through the collection

        # Arguments
        of: ``Callable`` Selector of reduce values
        reduce: Optional ``callable`` reduce method
        """
        if not callable(reduce):

            def reduce(x):
                return x

        selection = self.map(of)

        if len(self) == 1:
            return selection

        try:
            rows = min(len(e) for e in selection)
        except TypeError:
            return selection

        section = [
            reduce([element[row] for element in selection])
            for row in range(rows)
        ]

        return self.__class__(section)

    def _value_retriever(self, value):
        if self._use_as_callable(value):
            return value

        return lambda item: data_get(item, value)

    def _use_as_callable(self, value):
        return not isinstance(value, basestring) and callable(value)

    def _set_items(self, items):
        self._items = items

    def _get_items(self, items):
        if isinstance(items, list):
            return items
        elif isinstance(items, tuple):
            return list(items)
        elif (
            isinstance(items, Collection)
            or str(type(items))
            == "<class 'backpack.collections.base_collection.BaseCollection'>"
        ):
            return items.all()
        elif hasattr("items", "to_list"):
            return items.to_list()

        return [items]

    def __len__(self):
        return len(self.items)

    def __iter__(self):
        yield from self.items

    def __getitem__(self, item):
        if isinstance(item, slice):
            return self.__class__.make(self.items[item])

        return self.items[item]

    def __setitem__(self, key, value):
        self.items[key] = value

    def __delitem__(self, key):
        del self.items[key]

    def __eq__(self, other):
        if isinstance(other, Collection):
            other = other.items

        return other == self.items

    def __ne__(self, other):
        if isinstance(other, Collection):
            other = other.items

        return other != self.items

    def __str__(self):
        if len(self.items) > 5:
            items = ", ".join([repr(item) for item in self.items[:2]])
            items += " ... "
            items += ", ".join([repr(item) for item in self.items[-2:]])
        else:
            items = ", ".join([repr(item) for item in self.items])
        return f"Collection ({len(self.items)}) <{items}>"

    def __repr__(self):
        return self.__str__()


def collect(elements):
    return Collection(elements)


class ElementCollection(Collection):
    def as_dataframe(self):
        """Returns collection as Pandas dataframe"""
        data = {k: [] for k in self._items[0].serialize().keys()}
        for item in self._items:
            for k, v in item.serialize().items():
                data[k].append(v)
        import pandas

        return pandas.DataFrame.from_dict(data)

    def filter_by_module(self, module):
        return self.filter(lambda x: x.module == module)

    def filter_by_context(
        self,
        module: str,
        version: VersionType = None,
        **kwargs,
    ):
        from machinable import Element

        instance = Element.make(module, version, **kwargs)

        return self.filter(lambda x: x.matches(instance.compute_context()))

    def singleton(
        self,
        module: str,
        version: VersionType = None,
        **kwargs,
    ) -> Union[Any, "Component"]:
        from machinable import Element

        instance = Element.make(module, version, **kwargs)
        context = instance.compute_context()

        for candidate in self:
            if candidate.matches(context) and not candidate.hidden():
                return candidate

        return instance

    def __str__(self):
        return f"Elements <{len(self.items)}>"


class InterfaceCollection(ElementCollection):
    pass


class ComponentCollection(InterfaceCollection):
    def __str__(self):
        if len(self.items) > 15:
            items = ", ".join([repr(item) for item in self.items[:5]])
            items += " ... "
            items += ", ".join([repr(item) for item in self.items[-5:]])
        else:
            items = ", ".join([repr(item) for item in self.items])
        return f"Components ({len(self.items)}) <{items}>"

    def launch(self) -> "ComponentCollection":
        """Executes all components in the collection"""
        for component in self:
            component.launch()

        return self


class ExecutionCollection(ElementCollection):
    def __str__(self):
        return f"Executions <{len(self.items)}>"

    def status(self, status="started"):
        """Filters the collection by a status attribute

        # Arguments
        status: String, status field: 'started', 'finished', 'alive'
        """
        try:
            return self.filter(lambda item: getattr(item, "is_" + status)())
        except AttributeError as _ex:
            raise ValueError(f"Invalid status field: {status}") from _ex



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import inspect
import shlex
import sys

from flatten_dict import flatten

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from typing import Callable, Literal

import os
from functools import partial

import dill as pickle
from machinable import errors, schema
from machinable.collection import (
    Collection,
    ComponentCollection,
    InterfaceCollection,
)
from machinable.element import _CONNECTIONS as connected_elements
from machinable.element import Element, get_dump, get_lineage
from machinable.types import VersionType
from machinable.utils import (
    id_from_uuid,
    is_directory_version,
    joinpath,
    load_file,
    object_hash,
    save_file,
    update_uuid_payload,
)
from uuid_extensions import uuid7


def cachable(
    memory: bool = True,
    file: bool = True,
    fail_mode: Literal["ignore", "raise", "warn"] = "ignore",
) -> Callable:
    if not memory and not file:
        raise ValueError(
            "At least one of memory or file cache must be enabled."
        )

    def _decorator(fn: Callable) -> Callable:
        def _wrapper(self, *args, **kwargs):
            if not self.cached():
                return fn(self, *args, **kwargs)
            try:
                key = object_hash(
                    {"fn": fn.__name__, "args": args, "kwargs": kwargs}
                )[:8]
            except TypeError as _ex:
                if fail_mode == "raise":
                    raise
                else:
                    if fail_mode == "warn":
                        print(
                            f"Warning: Caching disabled for {fn.__name__} due to unhashable arguments. ({repr(_ex)})"
                        )
                    return fn(self, *args, **kwargs)

            fp = f".cachable_{key}.p"

            # in-memory cache
            if memory and fp in self._cache:
                return self._cache[fp]

            # file cache
            if file:
                miss = object()
                cached = self.load_file(fp, miss)

                if cached is miss:
                    result = fn(self, *args, **kwargs)
                    self.save_file(fp, result)
                    if memory:
                        self._cache[fp] = result
                    return result
            else:
                cached = fn(self, *args, **kwargs)

            if memory:
                self._cache[fp] = cached

            return cached

        return _wrapper

    return _decorator


class Relation:
    inverse: bool = False
    multiple: bool = False

    def __init__(
        self,
        fn,
        cached: bool = True,
        collection: Optional[Collection] = None,
        key: Optional[str] = None,
    ) -> None:
        self.fn = fn
        self.cached = cached
        self.collection = collection
        self.key = key

        self.cls = None
        self._related_cls = None

    @property
    def related_cls(self) -> "Interface":
        if self._related_cls is None:
            self._related_cls = self.fn()
        return self._related_cls

    @property
    def name(self) -> str:
        if not self.inverse:
            return f"{self.cls.kind}.{self.related_cls.kind}.{self.key or 'default'}"
        else:
            return f"{self.related_cls.kind}.{self.cls.kind}.{self.key or 'default'}"

    def collect(self, elements: List["Interface"]) -> Collection:
        if self.collection is None:
            return self.related_cls.collect(elements)
        return self.collection(elements)

    def __set_name__(self, cls, name):
        self.cls = cls
        if cls.__relations__ is None:
            cls.__relations__ = {}
        cls.__relations__[name] = self

    def __get__(self, instance, owner):
        if (
            not instance._relation_cache.get(self.fn.__name__, None)
            and instance.is_mounted()
        ):
            from machinable.index import Index

            related = Index.get().find_related(
                relation=self.name, uuid=instance.uuid, inverse=self.inverse
            )

            if related is not None:
                related = [
                    Interface.find_by_id(r.uuid, fetch=False) for r in related
                ]

                if self.multiple is False:
                    related = related[0] if len(related) > 0 else None
                else:
                    related = self.collect(related)

            instance._relation_cache[self.fn.__name__] = self.cached
            instance.__related__[self.fn.__name__] = related

        return instance.__related__[self.fn.__name__]


class HasOne(Relation):
    pass


class HasMany(Relation):
    multiple = True


class BelongsTo(Relation):
    inverse = True


class BelongsToMany(Relation):
    inverse = True
    multiple = True


def _relation(cls: Relation) -> Any:
    def _wrapper(
        f: Optional[Callable] = None,
        *,
        cached: bool = True,
        collection: Optional[Collection] = None,
        key: Optional[str] = None,
    ) -> Any:
        if f is None:
            return partial(cls, cached=cached, collection=collection, key=key)

        return cls(f, cached=cached, collection=collection, key=key)

    return _wrapper


belongs_to = _relation(BelongsTo)
has_one = _relation(HasOne)
has_many = _relation(HasMany)
belongs_to_many = _relation(BelongsToMany)


def _uuid_symlink(directory, uuid, mode=None):
    dst = os.path.join(directory, "related", id_from_uuid(uuid))
    try:
        if mode is None:
            return uuid

        os.makedirs(dst, exist_ok=True)
        if mode == "file":
            with open(os.path.join(dst, "link"), "w") as f:
                f.write("../../" + uuid)
        else:
            os.symlink("../../" + uuid, os.path.join(dst, "link"))
    except OSError:
        pass

    return uuid


class Interface(Element):
    kind = "Interface"
    default = None
    # class level relationship information
    # note that the actual data is kept in
    # the __related__ object propery and
    # existence should be checked on
    # actual interface instance
    __relations__: Optional[Dict[str, Relation]] = None

    def __init__(
        self,
        version: VersionType = None,
        uses: Union[None, "Interface", List["Interface"]] = None,
        derived_from: Optional["Interface"] = None,
    ):
        super().__init__(version=version)
        self.__model__ = schema.Interface(
            kind=self.kind,
            module=self.__model__.module,
            config=self.__model__.config,
            version=self.__model__.version,
            lineage=get_lineage(self),
        )
        self.__model__._dump = get_dump(self)

        self._kwargs["uses"] = uses
        self._kwargs["derived_from"] = derived_from

        # initialize relation data
        self.__related__ = {}
        self._relation_cache = {}
        if self.__relations__ is None:
            self.__relations__ = {}
        # relation_names = defaultdict(lambda: 0)
        for name, relation in self.__relations__.items():
            # if relation_names[relation.name] > 2:
            #     raise RuntimeError(f"Relationship name '{relation.name}' is ambigous. Set key on '{relation.fn.__name__}' to distinguish.")
            # relation_names[relation.name] += 1
            if relation.multiple:
                self.__related__[name] = relation.collect([])
            else:
                self.__related__[name] = None
        if uses:
            if not isinstance(uses, (list, tuple)):
                uses = [uses]
            for use in uses:
                self.__related__["uses"].append(use)
            self._relation_cache["uses"] = True

        if derived_from:
            self.__related__["ancestor"] = derived_from
            self._relation_cache["ancestor"] = True

        self._deferred_data = {}
        self._futures_stack = set()
        self._future_launch = False

    @classmethod
    def collect(cls, elements) -> InterfaceCollection:
        return InterfaceCollection(elements)

    def push_related(self, key: str, value: "Interface") -> None:
        if self.is_committed():
            raise errors.MachinableError(
                f"{repr(self)} already exists and cannot be modified."
            )
        if self.__relations__[key].multiple:
            self.__related__[key].append(value)
        else:
            self.__related__[key] = value
        self._relation_cache[key] = True

    def is_staged(self) -> bool:
        return self.__model__.uuid[-12:] != "0" * 12

    def stage(self) -> Self:
        self.__model__.context = context = self.compute_context()
        self.__model__.uuid = update_uuid_payload(self.__model__.uuid, context)

        # ensure that configuration and predicate has been computed
        assert self.config is not None
        self.__model__.predicate = self.compute_predicate()

        return self

    def is_committed(self) -> bool:
        from machinable.index import Index

        return Index.get().find_by_id(self.uuid) is not None

    def commit(self) -> Self:
        from machinable.index import Index

        index = Index.get()

        # only commit if not already in index
        if index.find_by_id(self.uuid) is not None:
            return self

        if self.on_before_commit() is False:
            # allow on_before_commit to abort
            return self

        if not self.is_staged():
            self.stage()

        if self.on_commit() is False:
            # allow on_commit to abort
            return self

        # commit to index
        for k, v in self.__related__.items():
            if v is None:
                continue
            r = self.__relations__[k]
            if not r.multiple:
                v = [v]
            for u in [i.commit().uuid for i in v]:
                if r.inverse:
                    index.create_relation(r.name, u, self.uuid)
                else:
                    index.create_relation(r.name, self.uuid, u)
        index.commit(self.__model__)

        # save to directory
        self.to_directory(self.local_directory(create=True))

        # commit to storage
        from machinable.storage import Storage

        for storage in Storage.connected():
            storage.commit(self)

        # write deferred files
        for filepath, data in self._deferred_data.items():
            self.save_file(filepath, data)
        self._deferred_data = {}

        self.on_after_commit()

        return self

    def on_before_commit(self):
        """Event hook before interface is committed."""

    def on_commit(self):
        """Event hook during interface commit when uuid, config, context
        and predicate have been computed and commit is about to be performed"""

    def on_after_commit(self):
        """Event hook after interface has been committed."""

    @belongs_to
    def project():
        from machinable.project import Project

        return Project

    @has_many(cached=False, key="derivatives")
    def derived() -> InterfaceCollection:
        """Returns a collection of derived interfaces"""
        return Interface

    @belongs_to(key="derivatives")
    def ancestor() -> Optional["Interface"]:
        """Returns parent interface or None if interface is independent"""
        return Interface

    @has_many(key="using")
    def uses() -> InterfaceCollection:
        return Interface

    @belongs_to_many(key="using")
    def used_by() -> InterfaceCollection:
        return Interface

    def related(self, deep: bool = False) -> InterfaceCollection:
        """Returns a collection of related interfaces"""
        collection = []
        for k, v in self.__relations__.items():
            if getattr(self, k, False) is False:
                continue
            if v.multiple:
                collection.extend(getattr(self, k))
            else:
                r = getattr(self, k)
                if r is not None:
                    collection.append(r)

        if deep:
            seen = {self.uuid}
            for i in collection:
                if not hasattr(i, "related") or i.uuid in seen:
                    continue
                collection.extend(i.related(deep=False).all())
                seen.add(i.uuid)

        return InterfaceCollection(collection).unique(lambda x: x.uuid)

    def related_iterator(self):
        seen = {self.uuid}
        for relation_attribute, relationship in self.__relations__.items():
            if getattr(self, relation_attribute, False) is False:
                continue
            related = getattr(self, relation_attribute)
            if not related:
                continue
            if not relationship.multiple:
                related = [related]
            for r in related:
                if r.uuid in seen:
                    continue
                yield r, relationship, seen
                seen.add(r.uuid)

    def to_cli(self) -> str:
        cli = [self.module]
        for v in self.__model__.version:
            if isinstance(v, str):
                cli.append(v)
            else:
                cli.extend(
                    [
                        f"{key}={shlex.quote(str(val))}"
                        for key, val in flatten(v, reducer="dot").items()
                    ]
                )

        return " ".join(cli)

    def derive(
        self,
        module: Union[str, Element, None] = None,
        version: VersionType = None,
        **kwargs,
    ) -> Self:
        if module is None:
            return self.make(module, version, derived_from=self, **kwargs)

        return self.derived.singleton(
            module, version, derived_from=self, **kwargs
        )

    @classmethod
    def singleton(
        cls,
        module: Union[str, "Element"],
        version: VersionType = None,
        **kwargs,
    ) -> "Collection":
        if module in [
            "machinable.index",
            "machinable.project",
        ] and is_directory_version(version):
            # interpret as shortcut for directory
            version = {"directory": version}
        candidates = cls.find(
            module,
            version,
            **kwargs,
        ).filter(lambda x: not x.hidden())

        if candidates:
            return candidates[-1]

        return cls.make(module, version, **kwargs)

    def is_mounted(self) -> bool:
        if self.__model__ is None:
            return False

        return os.path.exists(self.local_directory())

    @classmethod
    def find_by_id(cls, uuid: str, fetch: bool = True) -> Optional["Interface"]:
        from machinable.index import Index

        index = Index.get()

        model = index.find_by_id(uuid)
        if not model:
            return None

        if fetch is False and not model.module.startswith("__session__"):
            return cls.from_model(model)

        local_directory = index.local_directory(model.uuid)

        if not os.path.exists(local_directory):
            # try to fetch storage
            from machinable.storage import fetch as fetch_from_storage

            if not fetch_from_storage(model.uuid, local_directory):
                return None

        return cls.from_directory(local_directory)

    @classmethod
    def find_many_by_id(
        cls, uuids: List[str], fetch: bool = True
    ) -> "InterfaceCollection":
        return cls.collect([cls.find_by_id(uuid, fetch) for uuid in uuids])

    @classmethod
    def find_by_hash(
        cls, context_hash: str, fetch: bool = True
    ) -> Optional["Interface"]:
        from machinable.index import Index

        index = Index.get()

        return cls.collect(
            [
                cls.find_by_id(m.uuid, fetch)
                for m in index.find_by_hash(context_hash)
            ]
        )

    @classmethod
    def find(
        cls,
        module: Union[str, "Element"],
        version: VersionType = None,
        **kwargs,
    ) -> "InterfaceCollection":
        from machinable.index import Index

        index = Index.get()

        if module is None:
            context = {"predicate": {}}
            for scope in connected_elements["Scope"]:
                context["predicate"].update(scope())
            return cls.collect(
                [
                    cls.find_by_id(interface.uuid, fetch=False)
                    for interface in index.find_by_context(context)
                ]
            ).filter(lambda i: i.matches(context))

        try:
            candidate = cls.make(module, version, **kwargs)
        except ModuleNotFoundError:
            return cls.collect([])

        context = candidate.compute_context()
        if context is None:
            return cls.collect([])

        found = []

        for model in index.find_by_context(context):
            if not model.module.startswith("__session__"):
                found.append(cls.from_model(model))
                continue

            local_directory = index.local_directory(model.uuid)
            if not os.path.exists(local_directory):
                from machinable.storage import fetch as fetch_from_storage

                if not fetch_from_storage(model.uuid, local_directory):
                    continue

            # if in-session object, we need to update the pickled update
            #  since it might have changed
            assert not isinstance(module, str)
            save_file([local_directory, "dump.p"], pickle.dumps(module))

            found.append(cls.from_directory(local_directory))

        return cls.collect(found).filter(lambda i: i.matches(context))

    @classmethod
    def from_directory(cls, directory: str) -> Self:
        """Returns an interface from a storage directory

        Note that this does not verify the integrity of the directory.
        In particular, the interface may be missing or not be indexed.
        """
        data = load_file([directory, "model.json"])

        model = getattr(schema, data["kind"], None)
        if model is None:
            # TODO: users should have an option to register custom interface types
            raise ValueError(f"Invalid interface kind: {model['kind']}")

        interface = model(**data)
        if interface.module.startswith("__session__"):
            interface._dump = load_file([directory, "dump.p"], None)

        return cls.from_model(interface)

    def to_directory(self, directory: str, relations: bool = True) -> Self:
        save_file([directory, ".machinable"], self.__model__.uuid)
        save_file([directory, "model.json"], self.__model__)
        if self.__model__._dump is not None:
            save_file([directory, "dump.p"], self.__model__._dump)

        def _write_meta(d, r, uuid, related_uuid):
            save_file(
                [d, "related", "metadata.jsonl"],
                {
                    "uuid": uuid,
                    "related_uuid": related_uuid,
                    "name": r.name,
                    "multiple": r.multiple,
                    "inverse": r.inverse,
                    "cached": r.cached,
                    "fn": r.fn.__name__,
                },
                mode="a",
            )

        if relations:
            for k, v in self.__related__.items():
                if not v:
                    continue
                r = self.__relations__[k]
                if not r.multiple:
                    v = [v]
                # forward
                save_file(
                    [directory, "related", k],
                    "\n".join([_uuid_symlink(directory, i.uuid) for i in v])
                    + "\n",
                    mode="a",
                )
                for u in v:
                    if r.inverse:
                        _write_meta(directory, r, u.uuid, self.uuid)
                    else:
                        _write_meta(directory, r, self.uuid, u.uuid)
                # inverse
                for i in v:
                    try:
                        ir = [
                            _
                            for _ in i.__relations__.values()
                            if _.name == r.name and _ is not r
                        ][0]
                        save_file(
                            [i.local_directory(), "related", ir.fn.__name__],
                            _uuid_symlink(i.local_directory(), self.uuid)
                            + "\n",
                            mode="a",
                        )
                        if ir.inverse:
                            _write_meta(
                                i.local_directory(), ir, i.uuid, self.uuid
                            )
                        else:
                            _write_meta(
                                i.local_directory(), ir, self.uuid, i.uuid
                            )
                    except:
                        pass

        return self

    def fetch(
        self, directory: Optional[str] = None, force: bool = False
    ) -> bool:
        if not self.is_committed():
            return False

        if "fetched" in self._cache and not force:
            return True

        if directory is None:
            from machinable.index import Index

            directory = Index.get().local_directory(self.uuid)

        if not os.path.exists(directory) or force:
            from machinable.storage import fetch

            if not fetch(self.uuid, directory):
                return False

        self._cache["fetched"] = True
        return True

    def local_directory(self, *append: str, create: bool = False) -> str:
        from machinable.index import Index

        directory = Index.get().local_directory(self.uuid, *append)

        if create:
            os.makedirs(directory, exist_ok=True)

        return directory

    def load_file(
        self, filepath: Union[str, List[str]], default=None
    ) -> Optional[Any]:
        filepath = joinpath(filepath)
        if not self.is_mounted():
            # has write been deferred?
            if filepath in self._deferred_data:
                return self._deferred_data[filepath]

            return default

        data = load_file(self.local_directory(filepath), default=None)

        return data if data is not None else default

    def save_file(self, filepath: Union[str, List[str]], data: Any) -> str:
        filepath = joinpath(filepath)

        if os.path.isabs(filepath):
            raise ValueError("Filepath must be relative")

        if not self.is_mounted():
            # defer writes until interface storage is mounted
            self._deferred_data[filepath] = data
            return "$deferred"

        file = save_file(self.local_directory(filepath), data, makedirs=True)

        return file

    def launch(self) -> Self:
        ...

    @property
    def components(self) -> "ComponentCollection":
        if "components" not in self._cache:
            from machinable.execution import Execution

            with Execution().deferred() as e:
                self.launch()
            self._cache["components"] = e.executables

        return self._cache["components"]

    def cached(self):
        return self.components.reduce(
            lambda result, x: result and x.cached(), True
        )

    def future(self) -> Optional[Self]:
        from machinable.execution import Execution

        if Execution.is_connected():
            if self._future_launch is True:
                # prevent recursion
                return None

            try:
                self._future_launch = True
                self.launch()
            finally:
                self._future_launch = False

            return None

        ready = self.cached()

        # if this is called within an interface, we keep
        #  track of the state for later reference
        outer = None
        stack = inspect.stack()
        try:
            outer = stack[1][0].f_locals.get("self", None)
        finally:
            del stack

        if isinstance(outer, Interface) and outer != self:
            if ready:
                outer._futures_stack.discard(self.id)
            else:
                outer._futures_stack.add(self.id)

        if not ready or len(self._futures_stack) > 0:
            return None

        return self

    def hidden(
        self, hidden: Optional[bool] = None, reason: str = "user"
    ) -> bool:
        if hidden is None:
            return self.load_file("hidden", None) is not None
        elif hidden is True:
            self.save_file("hidden", str(reason))
            return True
        elif hidden is False:
            try:
                os.remove(self.local_directory("hidden"))
            except OSError:
                pass

        return hidden

    # a posteriori modifiers

    def all(self) -> "InterfaceCollection":
        module = (
            self.module
            if not self.module.startswith("__session__")
            else self.__class__
        )
        return self.find(module, self.__model__.version, **self._kwargs)

    def new(self) -> Self:
        return self.make(self.module, self.__model__.version, **self._kwargs)



from typing import TYPE_CHECKING, List, Optional, Union

import inspect
import os
import random
import sys
import threading

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from typing import Dict

from machinable import errors, schema
from machinable.collection import ComponentCollection, ExecutionCollection
from machinable.element import _CONNECTIONS as connected_elements
from machinable.element import get_dump, get_lineage
from machinable.index import Index
from machinable.interface import Interface, belongs_to, belongs_to_many
from machinable.project import Project
from machinable.storage import Storage
from machinable.types import VersionType
from machinable.utils import generate_seed

if TYPE_CHECKING:
    from machinable.execution import Execution


class Component(Interface):
    kind = "Component"
    default = None

    def __init__(
        self,
        version: VersionType = None,
        uses: Union[None, "Interface", List["Interface"]] = None,
        derived_from: Optional["Interface"] = None,
        seed: Union[int, None] = None,
    ):
        super().__init__(version=version, uses=uses, derived_from=derived_from)
        if seed is None:
            seed = generate_seed()
        self.__model__ = schema.Component(
            kind=self.kind,
            module=self.__model__.module,
            config=self.__model__.config,
            version=self.__model__.version,
            seed=seed,
            lineage=get_lineage(self),
        )
        self.__model__._dump = get_dump(self)
        self._current_execution_context = None

    @property
    def current_execution_context(self) -> "Execution":
        if self._current_execution_context is None:
            from machinable.execution import Execution

            self._current_execution_context = Execution.get()
        return self._current_execution_context

    @belongs_to_many(key="execution_history")
    def executions() -> ExecutionCollection:
        from machinable.execution import Execution

        return Execution

    @property
    def components(self) -> "ComponentCollection":
        return ComponentCollection([self])

    @property
    def execution(self) -> "Execution":
        from machinable.execution import Execution

        related = None
        if self.is_mounted():
            # if mounted, search for related, most recent execution
            related = Index.get().find_related(
                relation="Execution.Component.execution_history",
                uuid=self.uuid,
                inverse=True,
            )

            if related is not None and len(related) > 0:
                related = Interface.find_by_id(
                    sorted(related, key=lambda x: x.timestamp, reverse=True)[
                        0
                    ].uuid
                )
            else:
                related = None

        # use context if no related execution was found
        if related is None:
            if Execution.is_connected():
                related = Execution.get()
            else:
                related = self.current_execution_context

        related.of(self)

        return related

    def launch(self) -> Self:
        from machinable.execution import Execution

        self.fetch()

        if Execution.is_connected():
            # stage only, defer execution
            if not self.is_staged():
                self.stage()
            Execution.get().add(self)
        else:
            self.current_execution_context.add(self)
            self.current_execution_context.dispatch()

        return self

    @property
    def seed(self) -> int:
        return self.__model__.seed

    @property
    def nickname(self) -> str:
        return self.__model__.nickname

    @classmethod
    def collect(cls, components) -> "ComponentCollection":
        return ComponentCollection(components)

    def dispatch(self) -> Self:
        """Dispatch the component lifecycle"""
        writes_meta_data = (
            self.on_write_meta_data() is not False and self.is_mounted()
        )
        try:
            self.on_before_dispatch()

            self.on_seeding()

            # meta-data
            if writes_meta_data:
                if not self.execution.is_started():
                    self.execution.update_status(status="started")
                else:
                    self.execution.update_status(status="resumed")

                self.execution.save_file(
                    [self.id, "host.json"],
                    data=Project.get().provider().get_host_info(),
                )

            def beat():
                t = threading.Timer(15, beat)
                t.daemon = True
                t.start()
                self.on_heartbeat()
                if self.on_write_meta_data() is not False and self.is_mounted():
                    self.execution.update_status(status="heartbeat")
                return t

            heartbeat = beat()

            self.__call__()

            self.on_success()
            self.on_finish(success=True)

            if heartbeat is not None:
                heartbeat.cancel()

            if writes_meta_data:
                self.execution.update_status(status="finished")
                self.cached(True, reason="finished")

            self.on_after_dispatch(success=True)
        except BaseException as _ex:  # pylint: disable=broad-except
            self.on_failure(exception=_ex)
            self.on_finish(success=False)
            self.on_after_dispatch(success=False)
            raise errors.ComponentException(
                f"{self.__class__.__name__} dispatch failed"
            ) from _ex
        finally:
            if writes_meta_data:
                # propagate changes
                for storage in Storage.connected():
                    storage.update(self)

    def cached(
        self, cached: Optional[bool] = None, reason: str = "user"
    ) -> bool:
        if cached is None:
            return self.load_file("cached", None) is not None
        elif cached is True:
            self.save_file("cached", str(reason))
            return True
        elif cached is False:
            try:
                os.remove(self.local_directory("cached"))
            except OSError:
                pass

        return cached

    def dispatch_code(
        self,
        inline: bool = True,
        project_directory: Optional[str] = None,
        python: Optional[str] = None,
    ) -> Optional[str]:
        if project_directory is None:
            project_directory = Project.get().path()
        if python is None:
            python = sys.executable
        lines = ["from machinable import Project, Element, Component"]
        # context
        lines.append(f"Project('{project_directory}').__enter__()")
        for kind, elements in connected_elements.items():
            if kind in ["Project", "Execution"]:
                continue
            for element in elements:
                jn = element.as_json().replace('"', '\\"').replace("'", "\\'")
                lines.append(f"Element.from_json('{jn}').__enter__()")
        # dispatch
        lines.append(
            f"component__ = Component.from_directory('{os.path.abspath(self.local_directory())}')"
        )
        lines.append("component__.dispatch()")

        if inline:
            code = ";".join(lines)
            return f'{python} -c "{code}"'

        return "\n".join(lines)

    # life cycle

    def __call__(self) -> None:
        ...

    def on_before_dispatch(self) -> Optional[bool]:
        """Event triggered before the dispatch of the component"""

    def on_success(self):
        """Lifecycle event triggered iff execution finishes successfully"""

    def on_finish(self, success: bool):
        """Lifecycle event triggered right before the end of the component execution

        # Arguments
        success: Whether the execution finished sucessfully
        """

    def on_failure(self, exception: Exception) -> None:
        """Lifecycle event triggered iff the execution finished with an exception

        # Arguments
        exception: Execution exception
        """

    def on_after_dispatch(self, success: bool):
        """Lifecycle event triggered at the end of the dispatch.

        This is triggered independent of whether the execution has been successful or not.

        # Arguments
        success: Whether the execution finished sucessfully
        """

    def on_seeding(self):
        """Lifecycle event to implement custom seeding using `self.seed`"""
        random.seed(self.seed)

    def on_write_meta_data(self) -> Optional[bool]:
        """Event triggered before meta-data such as creation time etc. is written to the storage

        Return False to prevent writing of meta-data
        """

    def on_heartbeat(self) -> None:
        """Event triggered on heartbeat every 15 seconds"""



"""machinable"""
__all__ = [
    "Element",
    "Interface",
    "Execution",
    "Component",
    "Project",
    "Storage",
    "Mixin",
    "Index",
    "Scope",
    "mixin",
    "Schedule",
    "get",
]
__doc__ = """A modular system for machinable research code"""

from importlib import metadata as importlib_metadata

from machinable.cli import from_cli
from machinable.component import Component
from machinable.element import Element
from machinable.execution import Execution
from machinable.index import Index
from machinable.interface import Interface
from machinable.mixin import Mixin, mixin
from machinable.project import Project
from machinable.query import Query
from machinable.schedule import Schedule
from machinable.scope import Scope
from machinable.storage import Storage


def get_version() -> str:
    try:
        return importlib_metadata.version(__name__)
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return "unknown"


__version__: str = get_version()

get = Query()



from machinable import schema
from machinable.element import get_dump, get_lineage
from machinable.interface import Interface
from machinable.types import VersionType


class Schedule(Interface):
    """Schedule base class"""

    kind = "Schedule"

    def __init__(self, version: VersionType = None):
        super().__init__(version)
        self.__model__ = schema.Schedule(
            kind=self.kind,
            module=self.__model__.module,
            config=self.__model__.config,
            version=self.__model__.version,
            lineage=get_lineage(self),
        )
        self.__model__._dump = get_dump(self)



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

import getpass
import importlib
import os
import platform
import shutil
import socket
import subprocess
import sys
import urllib.request

import machinable
from machinable import schema
from machinable.element import (
    Element,
    extend,
    get_lineage,
    instantiate,
    normversion,
)
from machinable.errors import ConfigurationError
from machinable.interface import Interface
from machinable.types import VersionType
from machinable.utils import (
    find_subclass_in_module,
    get_commit,
    get_diff,
    get_root_commit,
    import_from_directory,
    is_directory_version,
)
from pydantic import BaseModel, Field, field_validator

if TYPE_CHECKING:
    from machinable.project import Project


def fetch_link(source, target):
    os.symlink(source, target, target_is_directory=True)
    return True


def fetch_directory(source, target):
    if not os.path.isdir(source):
        raise ValueError(f"{source} is not a directory")

    # we use symlink rather than copy for performance and consistency
    fetch_link(source, target)

    return True


def fetch_git(source: str, target: str) -> bool:
    name, directory = os.path.split(target)
    command = ["git", "clone", source, name]
    process = subprocess.Popen(command, cwd=directory)
    process.communicate()

    if process.returncode != 0:
        return False

    return True


def fetch_vendor(source, target):
    if source is None:
        return False

    # git
    if source.startswith("git+"):
        return fetch_git(source[4:], target)

    # symlink
    if source.startswith("link+"):
        return fetch_link(source[5:], target)

    # default: directory
    return fetch_directory(source, target)


def fetch_vendors(project: "Project"):
    top_level_vendor = project.get_root().path("vendor")
    os.makedirs(top_level_vendor, exist_ok=True)

    vendors = []
    for dependency in project.get_vendors():
        if isinstance(dependency, str):
            dependency = {dependency: None}

        for name, args in dependency.items():
            vendors.append(name)

            # source config
            if not isinstance(args, str):
                args = None

            # give local .machinablerc priority over project config
            source = args
            if isinstance(source, str):
                source = os.path.expanduser(source)

            # local target folder
            os.makedirs(project.path("vendor"), exist_ok=True)
            target = project.path("vendor", name)

            # protect against invalid symlinks
            if os.path.islink(target) and not os.path.exists(target):
                os.unlink(target)

            # skip if already existing
            if os.path.exists(target):
                break

            # top-level target folder
            top_level = os.path.join(top_level_vendor, name)

            # protect against invalid symlinks
            if os.path.islink(top_level) and not os.path.exists(top_level):
                os.unlink(top_level)

            if project.on_resolve_vendor(name, source, top_level) is not False:
                # fetch import to the top-level if it does not exist
                if not os.path.exists(top_level):
                    print(f"Fetching '+.{name}' to {top_level}")
                    if not fetch_vendor(source, top_level):
                        raise FileNotFoundError(
                            f"Could not fetch '+.{name}'. Please place it into {top_level}"
                        )

            # symlink from top-level to local target if not identical
            if top_level != target:
                os.symlink(top_level, target, target_is_directory=True)

            break

    return vendors


def import_element(
    directory: str, module: str, base_class: Any = None
) -> "Element":
    if base_class is None:
        base_class = Element

    # import project-relative
    module = (
        import_from_directory(
            module,
            directory,
            or_fail=False,
        )
        or module
    )

    # import globally
    if isinstance(module, str):
        try:
            module = importlib.import_module(module)
        except ModuleNotFoundError as _e:
            raise ModuleNotFoundError(
                f"Could not find module '{module}' in project {directory}"
            ) from _e

    element_class = find_subclass_in_module(module, base_class)
    if element_class is None:
        raise ConfigurationError(
            f"Could not find an element inheriting from the {base_class.__name__} base class. "
            f"Is it correctly defined in {module.__name__}?"
        )

    return element_class


class Project(Interface):
    kind = "Project"

    class Config(BaseModel):
        directory: Optional[str] = Field(default_factory=lambda: os.getcwd())

        @field_validator("directory")
        def normalize_directory(cls, v):
            return os.path.normpath(os.path.abspath(os.path.expanduser(v)))

    def __init__(
        self,
        version: VersionType = None,
    ):
        if is_directory_version(version):
            # interpret as shortcut for directory
            version = {"directory": version}
        super().__init__(version=version)
        self.__model__ = schema.Project(
            kind=self.kind,
            version=self.__model__.version,
            lineage=get_lineage(self),
        )
        self._parent: Optional[Project] = None
        self._provider: str = "interface/project"
        self._resolved_provider: Optional[Project] = None

    def provider(self, reload: Union[str, bool] = False) -> "Project":
        """Resolves and returns the provider instance"""
        if isinstance(reload, str):
            self._provider = reload
            self._resolved_provider = None
        if self._resolved_provider is None or reload:
            if isinstance(self._provider, str):
                self._resolved_provider = find_subclass_in_module(
                    module=import_from_directory(
                        self._provider, self.config.directory
                    ),
                    base_class=Project,
                    default=Project,
                )(version=self.__model__.version)
            else:
                self._resolved_provider = Project(
                    version=self.__model__.version
                )

        return self._resolved_provider

    @property
    def module(self) -> Optional[str]:
        if self.provider().__module__ != "machinable.project":
            return self._provider.replace("/", ".")

        return self.provider().__module__

    def add_to_path(self) -> None:
        if (
            os.path.exists(self.config.directory)
            and self.config.directory not in sys.path
        ):
            if self.is_root():
                sys.path.insert(0, self.config.directory)
            else:
                sys.path.append(self.config.directory)

    def name(self) -> str:
        return os.path.basename(self.config.directory)

    def path(self, *append: str) -> str:
        return os.path.join(self.config.directory, *append)

    def is_root(self) -> bool:
        return self._parent is None

    def get_root(self) -> "Project":
        if self.is_root():
            return self

        p = self._parent
        while not p.is_root():
            p = p._parent

        return p

    def get_vendors(self) -> List[str]:
        try:
            return [
                vendor
                for vendor in os.listdir(self.path("vendor"))
                if os.path.isdir(self.path("vendor", vendor))
                # todo: how to ignore vendors?
            ]
        except FileNotFoundError:
            return []

    def get_code_version(self) -> dict:
        return {
            "id": get_root_commit(self.config.directory),
            "project": get_commit(self.config.directory),
            "vendor": {
                vendor: get_commit(self.path("vendor", vendor))
                for vendor in self.get_vendors()
            },
        }

    def element(
        self,
        module: Union[str, Element],
        version: VersionType = None,
        base_class: Any = None,
        **constructor_kwargs,
    ) -> "Element":
        module, element_class = self.provider().on_resolve_element(module)

        module, version = extend(module, version)

        if not isinstance(module, str):
            # interactive session element
            element_class = module
            module = "__session__" + element_class.__name__
        else:
            # import from project
            if element_class is None:
                element_class = import_element(self.path(), module, base_class)

        element = instantiate(
            module,
            element_class,
            version,
            **constructor_kwargs,
        )

        return element

    def get_diff(self) -> Union[str, None]:
        return get_diff(self.path())

    def exists(self) -> bool:
        return os.path.exists(self.config.directory)

    def serialize(self) -> dict:
        return {
            "directory": self.config.directory,
        }

    @classmethod
    def unserialize(cls, serialized: dict) -> "Project":
        return cls(serialized)

    def get_host_info(self) -> dict:
        """Returned dictionary will be recorded as host information"""
        return {
            "network_name": platform.node(),
            "hostname": socket.gethostname(),
            "machine": platform.machine(),
            "python_version": platform.python_version(),
            "user": getpass.getuser(),
            "argv": sys.argv,
            "machinable_version": machinable.get_version(),
        }

    def resolve_remotes(self, module: str) -> Optional[Element]:
        remotes = self.on_resolve_remotes()

        if module not in remotes:
            return None

        remote = remotes[module]

        if isinstance(remote, (list, tuple)):
            remote, *dependencies = remote
            for dependency in dependencies:
                self.resolve_remotes(dependency)

        directory = self.path("interface/remotes")
        remote_module = module.replace(".", "_")
        filename = os.path.join(directory, remote_module + ".py")

        # obtain from remote if not existing
        if not os.path.exists(filename):
            os.makedirs(directory, exist_ok=True)

            if remote.startswith("url+"):
                # download
                with urllib.request.urlopen(remote[4:]) as response, open(
                    filename, "wb"
                ) as out_file:
                    data = response.read()
                    out_file.write(data)
            elif remote.startswith("link+"):
                # symlink
                try:
                    if os.path.islink(filename):
                        os.remove(filename)
                    os.symlink(remote[5:], filename)
                    print(remote[5:])
                except OSError:
                    pass
            elif remote.startswith("file+"):
                # copy
                shutil.copy(remote[5:], filename)
            else:
                raise ValueError(
                    f"Unknown remote type for {module}: {remote} (target file is {filename}"
                )

        try:
            element_class = find_subclass_in_module(
                import_from_directory(remote_module, directory, or_fail=True),
                Element,
            )
            if element_class is None:
                raise ModuleNotFoundError(
                    "Could not find an element in remote module"
                )
            return element_class
        except ModuleNotFoundError as _ex:
            raise ValueError(
                f"Invalid remote import: {remote} for {module} at {filename}"
            ) from _ex

    def on_resolve_remotes(self) -> Dict[str, str]:
        """Event triggered during remote resolution

        Return a dictionary of module names and their remote source.
        """
        return {}

    def on_resolve_element(
        self, module: Union[str, Element]
    ) -> Tuple[Union[str, Element], Optional[Element]]:
        """Override element resolution

        Return altered module and/or resolved Element class to be used instead.
        """
        if isinstance(module, str):
            try:
                return self._cache["on_resolve_element"][module]
            except KeyError:
                self._cache.setdefault("on_resolve_element", {})
                v = module, self.resolve_remotes(module)
                self._cache["on_resolve_element"][module] = v
                return v

        return module, None

    def on_resolve_vendor(
        self, name: str, source: str, target: str
    ) -> Optional[bool]:
        """Event triggered when vendor is resolved

        # Arguments
        name: The name of the vendor
        source: The source configuration
        target: The target directory (may or may not exists yet)

        Return False to prevent the default automatic resolution
        """

    def __repr__(self) -> str:
        return f"Project({self.config.directory})"



from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

import time

from machinable.utils import (
    empty_uuid,
    generate_nickname,
    generate_seed,
    id_from_uuid,
)
from pydantic import BaseModel, Field, PrivateAttr
from uuid_extensions.uuid7 import timestamp_ns


class Element(BaseModel):
    uuid: str = Field(default_factory=empty_uuid)
    kind: str = "Element"
    module: Optional[str] = None
    version: List[Union[str, Dict]] = []
    config: Optional[Dict] = None
    predicate: Optional[Dict] = None
    context: Optional[Dict] = None
    lineage: Tuple[str, ...] = ()

    @property
    def timestamp(self) -> int:
        try:
            return timestamp_ns(self.uuid, suppress_error=False)
        except ValueError:
            return 0

    @property
    def hash(self) -> str:
        return self.uuid[-12:]

    @property
    def id(self) -> str:
        return id_from_uuid(self.uuid)

    def extra(self) -> Dict:
        return {}


class Storage(Element):
    kind: str = "Storage"


class Index(Element):
    kind: str = "Index"


class Scope(Element):
    kind: str = "Scope"


class Interface(Element):
    kind: str = "Interface"
    _dump: Optional[bytes] = PrivateAttr(default=None)


class Component(Interface):
    kind: str = "Component"
    seed: int = Field(default_factory=generate_seed)
    nickname: str = Field(default_factory=generate_nickname)

    def extra(self) -> Dict:
        return {"seed": self.seed, "nickname": self.nickname}


class Project(Interface):
    kind: str = "Project"


class Execution(Interface):
    kind: str = "Execution"
    seed: int = Field(default_factory=generate_seed)
    resources: Optional[Dict] = None

    def extra(self) -> Dict:
        return {"seed": self.seed, "resources": self.resources}


class Schedule(Interface):
    kind: str = "Schedule"



__all__ = ["to_dict"]
from typing import TYPE_CHECKING, Any, Optional, Tuple, Union

import collections
import re
import warnings
from dataclasses import asdict, dataclass, is_dataclass
from inspect import isclass

import omegaconf
from pydantic import BaseModel

if TYPE_CHECKING:
    from machinable.element import Element


class _ModelPrototype(BaseModel):
    pass


def from_element(element: "Element") -> Tuple[dict, Optional[BaseModel]]:
    if not isclass(element):
        element = element.__class__

    if not hasattr(element, "Config"):
        return {}, None

    config = getattr(element, "Config")

    # free-form
    if isinstance(config, collections.abc.Mapping):
        return config, None

    # pydantic model
    if isinstance(config, type(_ModelPrototype)):
        with warnings.catch_warnings(record=True):  # ignore pydantic warnings
            return config().model_dump(), config

    # ordinary class
    if not is_dataclass(config):
        config = dataclass(config)

    # dataclass
    return asdict(config()), None


def match_method(definition: str) -> Optional[Tuple[str, str]]:
    fn_match = re.match(r"^(?P<method>\w+)\s?\((?P<args>[^()]*)\)$", definition)
    if fn_match is None:
        return None

    function = fn_match.groupdict()["method"]
    args = fn_match.groupdict()["args"]
    return function, args


def rewrite_config_methods(
    config: Union[collections.abc.Mapping, str, list, tuple]
) -> Any:
    if isinstance(config, list):
        return [rewrite_config_methods(v) for v in config]

    if isinstance(config, tuple):
        return (rewrite_config_methods(v) for v in config)

    if isinstance(config, collections.abc.Mapping):
        return {k: rewrite_config_methods(v) for k, v in config.items()}

    if isinstance(config, str) and match_method(config):
        # todo: take advatage of walrus operator in Python 3.8
        function, args = match_method(config)
        return '${config_method:"' + function + '","' + args + '"}'

    return config


def to_dict(dict_like):
    if isinstance(dict_like, (omegaconf.DictConfig, omegaconf.ListConfig)):
        return omegaconf.OmegaConf.to_container(dict_like)

    if isinstance(dict_like, (list, tuple)):
        return dict_like.__class__([k for k in dict_like])

    if not isinstance(dict_like, collections.abc.Mapping):
        return dict_like

    return {k: to_dict(v) for k, v in dict_like.items()}



from typing import TYPE_CHECKING, List, Optional

import os
import sys

from omegaconf import OmegaConf

if TYPE_CHECKING:
    from machinable.types import VersionType


def parse(args: List) -> tuple:
    kwargs = []
    methods = []
    elements = []
    dotlist = []
    version = []

    def _parse_dotlist():
        if len(dotlist) == 0:
            return
        _ver = {}
        for k, v in OmegaConf.to_container(
            OmegaConf.from_dotlist(dotlist)
        ).items():
            if k.startswith("**"):
                kwargs[-1][k[2:]] = v
            else:
                _ver[k] = v
        version.append(_ver)

    def _push():
        if len(version) == 0:
            return

        if len(elements) > 0:
            elements[-1].extend(version)
        else:
            elements.append(version)

    for arg in args:
        if arg.startswith("~"):
            # version
            _parse_dotlist()
            dotlist = []
            version.append(arg)
        elif arg.startswith("--"):
            # method
            methods.append((len(elements), arg[2:]))
        elif "=" in arg:
            # dotlist
            dotlist.append(arg)
        else:
            # module
            _parse_dotlist()
            _push()
            dotlist = []
            version = []
            # auto-complete `.project` -> `interface.project`
            if arg.startswith("."):
                arg = "interface" + arg
            elements.append([arg])
            kwargs.append({})

    _parse_dotlist()
    _push()

    return elements, kwargs, methods


def from_cli(args: Optional[List] = None) -> "VersionType":
    if args is None:
        args = sys.argv[1:]

    elements, _, _ = parse(args)

    return sum(elements, [])


def main(args: Optional[List] = None):
    import machinable

    if args is None:
        args = sys.argv[1:]

    if len(args) == 0:
        print("\nhelp")
        print("\nversion")
        print("\nget")
        return 0

    action, args = args[0], args[1:]

    if action == "help":
        h = "get"
        if len(args) > 0:
            h = args[0]

        if h == "get":
            print("\nmachinable get [element_module...] [version...] --method")
            print("\nExample:")
            print(
                "\tmachinable get my_component ~ver arg=1 nested.arg=2 --launch\n"
            )
            return 0
        elif h == "version":
            print("\nmachinable version")
            return 0
        else:
            print("Unrecognized option")
            return 128

    if action == "version":
        version = machinable.get_version()
        print(version)
        return 0

    if action.startswith("get"):
        sys.path.append(os.getcwd())

        get = machinable.get
        if action != "get":
            get = getattr(get, action.split(".")[-1])

        elements, kwargs, methods = parse(args)
        contexts = []
        component = None
        for i, (module, *version) in enumerate(elements):
            element = get(module, version, **kwargs[i])
            if i == len(elements) - 1:
                component = element
            else:
                contexts.append(element.__enter__())

        if component is None:
            raise ValueError("You have to provide at least one interface")

        for i, method in methods:
            # check if cli_{method} exists before falling back on {method}
            target = getattr(
                component, f"cli_{method}", getattr(component, method)
            )
            if callable(target):
                target()
            else:
                print(target)

        for context in reversed(contexts):
            context.__exit__()

        return 0

    print("Invalid argument")
    return 128



from typing import Any, Dict, List, Optional, Union

from arrow.arrow import Arrow

VersionType = Union[str, dict, None, List[Union[str, dict, None]]]
ElementType = List[Union[str, dict]]
DatetimeType = Arrow
TimestampType = Union[float, int, DatetimeType]
JsonableType = Dict[str, Union[str, float, int, None, DatetimeType]]
Optional[List[Union[str, dict]]]



import types
from types import ModuleType
from typing import Any, Callable, List, Mapping, Optional, Tuple, Union

import stat
import sys
import tokenize
from io import BytesIO

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

import hashlib
import importlib
import inspect
import json
import os
import random
import string
import subprocess
import sys
from collections import deque
from concurrent.futures import ThreadPoolExecutor
from importlib import metadata as importlib_metadata
from keyword import iskeyword
from pathlib import Path
from uuid import UUID

import arrow
import dill as pickle
import jsonlines
import omegaconf
from flatten_dict import unflatten as _unflatten_dict
from omegaconf.omegaconf import OmegaConf
from uuid_extensions import uuid7

sentinel = object()

import json

from machinable.types import DatetimeType, VersionType
from pydantic import BaseModel


def serialize(obj):
    """JSON serializer for objects not serializable by default json code"""
    if isinstance(obj, BaseModel):
        return obj.model_dump()
    if isinstance(obj, UUID):
        return obj.hex
    if isinstance(obj, DatetimeType):
        return str(obj)
    if isinstance(obj, (omegaconf.DictConfig, omegaconf.ListConfig)):
        return OmegaConf.to_container(obj)
    else:
        raise TypeError(f"Unserializable object {obj} of type {type(obj)}")


def normjson(
    data: Any, default: Optional[Callable[[Any], Any]] = serialize
) -> str:
    return json.dumps(
        data, sort_keys=True, separators=(",", ":"), default=default
    )


def generate_seed(random_state=None):
    """Generates a seed from a random state

    # Arguments
    random_state: Random state or None

    Returns:
    int32 seed value
    """
    if random_state is None or isinstance(random_state, int):
        random_state = random.Random(random_state)

    return random_state.randint(0, 2**31 - 1)


def timestamp_to_directory(timestamp: int) -> str:
    return (
        arrow.get(timestamp / 1e9)
        .strftime("%Y-%m-%dT%H%M%S_%f%z")
        .replace("+", "_")
    )


def is_valid_variable_name(name):
    if not isinstance(name, str):
        return False
    return name.isidentifier() and not iskeyword(name)


def is_valid_module_path(name):
    return all(map(is_valid_variable_name, name.split(".")))


def is_directory_version(version: VersionType) -> bool:
    if not isinstance(version, str):
        return False
    if not version.startswith("~"):
        return True
    if "/" in version or "\\" in version or version == "~":
        return True

    return False


def joinpath(filepath: Union[str, List[str]]) -> str:
    if isinstance(filepath, str):
        return filepath

    return os.path.join(*[str(p) for p in filepath if p is not None])


def random_str(length: int, random_state=None):
    if random_state is None or isinstance(random_state, int):
        random_state = random.Random(random_state)

    return "".join(
        random_state.choice(
            string.ascii_uppercase + string.ascii_lowercase + string.digits
        )
        for _ in range(length)
    )


def object_hash(
    payload: Any, default: Optional[Callable[[Any], Any]] = serialize
) -> str:
    json_str = normjson(payload, default=default)
    hash_obj = hashlib.sha256(json_str.encode())
    return hash_obj.hexdigest()


def file_hash(filepath: str):
    hash_obj = hashlib.blake2b()
    with open(str(filepath), "rb") as file:
        while chunk := file.read(8192):
            hash_obj.update(chunk)
    return hash_obj.hexdigest()[:12]


def update_uuid_payload(uuid: str, payload: dict) -> str:
    hash_hex = object_hash(payload)

    timestamp_part = uuid[:24]
    new_random_part = hash_hex[:12]

    return f"{timestamp_part}{new_random_part}"


def empty_uuid() -> str:
    uuid = uuid7(as_type="str")
    timestamp_part = uuid[:24]
    null_part = "0" * 12
    return timestamp_part + null_part


def id_from_uuid(uuid: str) -> str:
    return uuid[11:13] + uuid[14:18]


# These words are picked under two objectives:
#  (i) complex enough to make them memorizable
#  (ii) pronunciation should not pose uneccessary challenges
# Contributions of additions/modifications welcome!
_WORDS = {
    "tree": (
        "willow",
        "ivy",
        "lime",
        "tilia",
        "mapel",
        "oak",
        "elder",
        "cherry",
        "dogwood",
        "elm",
        "ash",
        "blackthorn",
        "fir",
        "crabapple",
        "beech",
        "birch",
        "salix",
        "juniper",
    ),
    "color": (
        "peridot",
        "tomato",
        "keppel",
        "melon",
        "pale",
        "zomp",
        "pastel",
        "lavender",
        "lapis",
        "ecru",
        "eggshell",
        "colbalt",
        "cerulean",
        "aero",
        "alabaster",
        "blush",
        "citrine",
        "chocolate",
        "coffee",
        "falu",
        "flax",
        "jet",
    ),
    "animal": (
        "albatross",
        "chinchilla",
        "alligator",
        "butterfly",
        "flamingo",
        "giraffe",
        "jellyfish",
        "mosquito",
        "raccoon",
        "weasel",
        "zebra",
        "hedgehog",
    ),
}


def generate_nickname(categories=None, glue="_"):
    """Generate a random nickname by chaining words from categories

    # Arguments
    categories: List of categories, available: 'animal', 'tree', 'color'.
                Can be nested. Defaults to ('color', ('animal', 'tree'))
    """
    if categories is None:
        categories = ("color", ("animal", "tree"))
    if isinstance(categories, str):
        categories = [categories]
    if not isinstance(categories, (list, tuple)):
        raise ValueError("Categories has to be a list of tuple")

    picks = []
    for category in categories:
        if isinstance(category, (list, tuple)):
            category = random.choice(category)
        if category not in _WORDS:
            raise KeyError(f"Invalid category: {category}")
        picks.append(random.choice(_WORDS[category]))
    return glue.join(picks)


def load_file(
    filepath: Union[str, List[str]],
    default: Any = sentinel,
    opener=open,
    **opener_kwargs,
) -> Any:
    """Loads a data object from file

    # Arguments
    filepath: Target filepath. The extension is being used to determine
        the file format. Supported formats are:
        .json (JSON), .jsonl (JSON-lines), .npy (numpy), .p (pickle); all other will be loaded as plain text (e.g. .txt|.log|.diff|.sh)
    default: Optional default if reading fails
    opener: Customer file opener
    opener_kwargs: Optional arguments to pass to the opener
    """
    filepath = joinpath(filepath)
    _, ext = os.path.splitext(filepath)
    mode = opener_kwargs.pop("mode", "r")
    try:
        if ext == ".p":
            if "b" not in mode:
                mode = mode + "b"
            with opener(filepath, mode, **opener_kwargs) as f:
                data = pickle.load(f)
        elif ext == ".json":
            with opener(filepath, mode, **opener_kwargs) as f:
                data = json.load(f)
        elif ext == ".jsonl":
            with jsonlines.Reader(
                opener(filepath, mode, **opener_kwargs)
            ) as reader:
                return list(reader.iter())
        elif ext == ".npy":
            import numpy as np

            if "b" not in mode:
                mode = mode + "b"
            with opener(filepath, mode, **opener_kwargs) as f:
                data = np.load(f, allow_pickle=True)
        else:
            with opener(filepath, mode, **opener_kwargs) as f:
                data = f.read()
        return data
    except (FileNotFoundError, Exception) as _ex:
        if default is not sentinel:
            return default
        raise _ex


def save_file(
    filepath: Union[str, List[str]],
    data: Any,
    makedirs: Union[bool, Callable] = True,
    opener=open,
    **opener_kwargs,
) -> str:
    """Saves a data object to file

    # Arguments
    filepath: Target filepath. The extension is being used to determine
        the file format. Supported formats are:
        .json (JSON), .jsonl (JSON-lines), .npy (numpy), .p (pickle); all other will be written as plain text (e.g. .txt|.log|.diff|.sh)
    data: The data object
    makedirs: If True or Callable, path will be created
    opener: Customer file opener
    opener_kwargs: Optional arguments to pass to the opener

    Returns the absolute path to the written file
    """
    filepath = joinpath(filepath)
    path = os.path.dirname(filepath)
    name = os.path.basename(filepath)
    _, ext = os.path.splitext(name)
    mode = opener_kwargs.pop("mode", "w")

    if path != "":
        if makedirs is True:
            os.makedirs(path, exist_ok=True)
        elif callable(makedirs):
            makedirs(path)

    if ext == ".json":
        # json
        with opener(filepath, mode, **opener_kwargs) as f:
            f.write(json.dumps(data, ensure_ascii=False, default=serialize))
    elif ext == ".jsonl":
        # jsonlines
        with jsonlines.Writer(
            opener(filepath, mode, **opener_kwargs),
            dumps=lambda *args, **kwargs: json.dumps(
                *args, sort_keys=True, default=serialize, **kwargs
            ),
        ) as writer:
            writer.write(data)
    elif ext == ".npy":
        import numpy as np

        if "b" not in mode:
            mode += "b"
        # numpy
        with opener(filepath, mode, **opener_kwargs) as f:
            np.save(f, data)
    elif ext == ".p":
        if "b" not in mode:
            mode += "b"
        with opener(filepath, mode, **opener_kwargs) as f:
            pickle.dump(data, f)
    else:
        # plain-text formats
        with opener(filepath, mode, **opener_kwargs) as f:
            f.write(str(data))

    return os.path.abspath(filepath)


def import_from_directory(
    module_name: str, directory: str, or_fail: bool = False
) -> Optional[ModuleType]:
    """Imports a module relative to a given absolute directory"""
    # determine the target .py file path
    file_path = os.path.join(directory, module_name.replace(".", "/"))
    tmp_file = False
    if os.path.isdir(file_path):
        file_path = os.path.join(file_path, "__init__.py")
        # if the __init__.py file does not exists, we temporily create it to emulate
        # the standard Python 3 import behaviour
        try:
            Path(file_path).touch(exist_ok=False)
            tmp_file = True
        except FileExistsError:
            pass
    else:
        file_path += ".py"

    try:
        spec = importlib.util.spec_from_file_location(module_name, file_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        sys.modules[module_name] = module
        return module
    except FileNotFoundError as _e:
        if or_fail:
            raise ModuleNotFoundError(
                f"No module named '{module_name}'"
            ) from _e
    finally:
        if tmp_file:
            os.remove(file_path)

    return None


def find_subclass_in_module(
    module: Optional[types.ModuleType],
    base_class: Any,
    default: Optional[Any] = None,
) -> Any:
    if module is not None:
        candidates = inspect.getmembers(
            module,
            lambda x: bool(
                # is class of given type that is defined in the target module
                inspect.isclass(x)
                and issubclass(x, base_class)
                and x.__module__ == module.__name__
            ),
        )
        if len(candidates) > 0:
            return candidates[0][1]

    return default


def find_installed_extensions(key: str) -> List[Tuple[str, ModuleType]]:
    return [
        (module.name, module.load())
        for module in importlib_metadata.entry_points().get(
            f"machinable.{key}", []
        )
    ]


def update_dict(
    d: Mapping, update: Optional[Mapping] = None, copy: bool = False
) -> Mapping:
    if d is None:
        d = {}
    if not isinstance(d, Mapping):
        raise ValueError(f"Expected mapping but found {type(d).__name__}: {d}")
    if copy:
        d = d.copy()
    if not update:
        return d
    if not isinstance(update, Mapping):
        if isinstance(update, str):
            raise ValueError(
                f"Invalid version {update}. Did you forget the ~-prefix?"
            )
        else:
            raise ValueError(
                f"Expected update mapping but found {type(update).__name__}: {update}"
            )
    for k, val in update.items():
        if isinstance(val, Mapping):
            d[k] = update_dict(d.get(k, {}), val, copy=copy)
        else:
            d[k] = val
    return d


def norm_version_call(version: str):
    code = version.replace("~", "").strip()
    tokens = tokenize.tokenize(BytesIO(code.encode("utf-8")).readline)
    normalized_tokens = []

    last_token_type = None
    for toknum, tokval, _, _, _ in tokens:
        if toknum in [
            tokenize.ENCODING,
            tokenize.ENDMARKER,
            tokenize.NEWLINE,
            tokenize.NL,
        ]:
            # skip encoding, endmarker, newlines
            continue
        if toknum == tokenize.STRING:
            # string literals are preserved
            normalized_tokens.append(tokval)
        elif toknum == tokenize.OP:
            # remove spaces before and after operators
            if (
                tokval == "="
                and normalized_tokens
                and normalized_tokens[-1] == " "
            ):
                normalized_tokens.pop()
            normalized_tokens.append(tokval)
        elif toknum in [tokenize.NAME, tokenize.NUMBER]:
            if last_token_type not in [
                None,
                tokenize.OP,
                tokenize.NL,
                tokenize.NEWLINE,
            ] and not (
                last_token_type == tokenize.OP
                and normalized_tokens[-1] in ["(", ","]
            ):
                normalized_tokens.append(" ")
            normalized_tokens.append(tokval)
        last_token_type = toknum

    if "~" in version:
        normalized_tokens = ["~"] + normalized_tokens

    normalized_code = "".join(normalized_tokens)
    normalized_code = (
        normalized_code.replace(" (", "(")
        .replace("( ", "(")
        .replace(" )", ")")
        .replace(", ", ",")
        .replace(" ,", ",")
    )
    return normalized_code


def dot_splitter(flat_key):
    if not isinstance(flat_key, str):
        raise ValueError(
            f"Expected string but found {type(flat_key).__name__}: {flat_key}"
        )
    return tuple(flat_key.split("."))


def unflatten_dict(
    d: Mapping,
    splitter: Union[str, Callable] = dot_splitter,
    inverse: bool = False,
    recursive: bool = True,
    copy: bool = True,
) -> dict:
    """Recursively unflatten a dict-like object

    # Arguments
    d: The dict-like to unflatten
    splitter: The key splitting method
        If a Callable is given, the Callable will be used to split `d`.
        'tuple': Use each element in the tuple key as the key of the unflattened dict.
        'path': Use `pathlib.Path.parts` to split keys.
        'underscore': Use underscores to split keys.
        'dot': Use underscores to split keys.
    inverse: If True, inverts the key and value before flattening
    recursive: Perform unflatting recursively
    copy: Creates copies to leave d unmodified
    """
    if recursive is False:
        return _unflatten_dict(d=d, splitter=splitter, inverse=inverse)

    if isinstance(d, list):
        return [
            unflatten_dict(v, splitter, inverse, recursive, copy) for v in d
        ]

    if isinstance(d, tuple):
        return (
            unflatten_dict(v, splitter, inverse, recursive, copy) for v in d
        )

    if isinstance(d, Mapping):
        if copy:
            d = d.copy()
        flat = _unflatten_dict(d=d, splitter=splitter, inverse=inverse)
        for k, nested in flat.items():
            flat[k] = unflatten_dict(nested, splitter, inverse, recursive, copy)
        return flat

    return d


def run_and_stream(
    args,
    *,
    stdout_handler: Callable = print,
    stderr_handler: Callable = print,
    check: bool = True,
    text: bool = True,
    env: Optional[dict] = sentinel,
    **kwargs,
) -> int:
    if env is sentinel:
        # workaround, see https://stackoverflow.com/a/60070753
        env = os.environ

    with subprocess.Popen(
        args,
        text=text,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env=env,
        **kwargs,
    ) as process:
        with ThreadPoolExecutor(2) as pool:

            def _st(stream, handler):
                # reduced memory iteration
                deque((handler(line) for line in stream), maxlen=0)

            pool.submit(_st, process.stdout, stdout_handler)
            pool.submit(_st, process.stderr, stderr_handler)
        exit_code = process.wait()
        if check and exit_code:
            raise subprocess.CalledProcessError(
                exit_code, process.args, process.stdout, process.stderr
            )
    return exit_code


def chmodx(filepath: str) -> str:
    st = os.stat(filepath)
    os.chmod(filepath, st.st_mode | stat.S_IEXEC)
    return filepath


def get_diff(repository: str) -> Optional[str]:
    command = ["git", "diff", "--staged"]
    try:
        process = subprocess.run(
            command,
            cwd=os.path.abspath(repository),
            capture_output=True,
            text=True,
            check=True,
        )
        return process.stdout
    except subprocess.CalledProcessError:
        return None


def get_commit(repository: str) -> dict:
    try:

        def run_git_command(command):
            return (
                subprocess.check_output(
                    ["git", "-C", repository] + command,
                    stderr=subprocess.DEVNULL,
                )
                .decode()
                .strip()
            )

        branch = run_git_command(["rev-parse", "--abbrev-ref", "HEAD"])
        path = run_git_command(["config", "--get", "remote.origin.url"])
        is_dirty = run_git_command(["status", "--porcelain"]) != ""
        commit = run_git_command(["rev-parse", "HEAD"])

        return {
            "path": path,
            "commit": commit,
            "is_dirty": is_dirty,
            "branch": branch,
        }
    except subprocess.CalledProcessError:
        pass

    return {"path": None, "commit": None, "is_dirty": None, "branch": None}


def get_root_commit(repository: str) -> Optional[str]:
    command = ["git", "rev-list", "--parents", "HEAD"]
    try:
        process = subprocess.run(
            command, cwd=repository, capture_output=True, text=True, check=True
        )
        return process.stdout.split("\n")[-2]
    except subprocess.CalledProcessError:
        return None


class Jsonable:
    def as_json(self, stringify=True, default=serialize, **dumps_kwargs):
        serialized = self.serialize()
        if stringify:
            serialized = json.dumps(serialized, default=default, **dumps_kwargs)
        return serialized

    @classmethod
    def from_json(cls, serialized, **loads_kwargs):
        if isinstance(serialized, str):
            serialized = json.loads(serialized, **loads_kwargs)
        return cls.unserialize(serialized)

    def clone(self):
        return self.__class__.from_json(self.as_json())

    def serialize(self) -> dict:
        raise NotImplementedError

    @classmethod
    def unserialize(cls, serialized: dict) -> Any:
        raise NotImplementedError


class Connectable:
    """Connectable trait"""

    __connection__: Optional["Connectable"] = None

    @classmethod
    def is_connected(cls) -> bool:
        return cls.__connection__ is not None

    @classmethod
    def get(cls) -> "Connectable":
        return cls() if cls.__connection__ is None else cls.__connection__

    def __enter__(self) -> Self:
        self._outer_connection = (  # pylint: disable=attribute-defined-outside-init
            self.__class__.__connection__
        )
        self.__class__.__connection__ = self
        return self

    def __exit__(self, *args, **kwargs) -> Self:
        if self.__class__.__connection__ is self:
            self.__class__.__connection__ = None
        if getattr(self, "_outer_connection", None) is not None:
            self.__class__.__connection__ = self._outer_connection
        return self



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

import collections
import copy
import json
import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

import inspect

import arrow
import dill as pickle
import machinable
import omegaconf
from machinable import schema
from machinable.collection import ElementCollection
from machinable.config import (
    from_element,
    match_method,
    rewrite_config_methods,
    to_dict,
)
from machinable.errors import ConfigurationError, MachinableError
from machinable.mixin import Mixin
from machinable.types import DatetimeType, ElementType, VersionType
from machinable.utils import (
    Jsonable,
    id_from_uuid,
    joinpath,
    norm_version_call,
    sentinel,
    serialize,
    unflatten_dict,
    update_dict,
)
from omegaconf import DictConfig, OmegaConf


class ConfigMethod:
    def __init__(self, element: "Element", prefix="config") -> None:
        self.element = element
        self.prefix = prefix

    def __call__(self, function, args) -> Any:
        from machinable.project import Project

        definition = f"{function}({args})"
        method = f"{self.prefix}_{function}"

        if hasattr(self.element, method):
            obj = "self.element."
        elif hasattr(Project.get().provider(), method):
            obj = "Project.get().provider()."
        else:
            raise ConfigurationError(
                f"{self.prefix.title()} method {definition} specified but {type(self.element).__name__}.{method}() does not exist."
            )

        # Using eval is evil, but in this case there is probably not enough at stake
        # to justify the implementation of a proper parser
        try:
            return eval(  # pylint: disable=eval-used
                obj + method + "(" + args + ")"
            )
        except Exception as _ex:
            raise ConfigurationError(
                f"{_ex} in {self.prefix} method {type(self).__name__}.{method}()"
            ) from _ex


def normversion(version: VersionType = None) -> List[Union[str, dict]]:
    if not isinstance(
        version,
        (
            list,
            tuple,
            omegaconf.listconfig.ListConfig,
        ),
    ):
        version = [version]

    def _valid(item):
        if item is None or item == {}:
            # skip
            return False

        if not isinstance(item, (collections.abc.Mapping, str)):
            raise ValueError(
                f"Invalid version. Expected str or dict but found {type(item).__name__}: {item}"
            )

        return True

    def _norm(item):
        if isinstance(item, collections.abc.Mapping):
            # convert to dict and unflatten
            return unflatten_dict(
                OmegaConf.to_container(OmegaConf.create(item)), recursive=False
            )
        if isinstance(item, str) and "~" in item:
            return norm_version_call(item)

        return item

    return [_norm(v) for v in version if _valid(v)]


def compact(
    element: Union[str, List[Union[str, dict, None]]],
    version: VersionType = None,
) -> ElementType:
    if isinstance(element, (list, tuple, omegaconf.listconfig.ListConfig)):
        element, *default_version = element
        if not isinstance(version, (list, tuple)):
            version = [version]
        version = list(default_version) + version

    if not isinstance(element, str):
        raise ValueError(
            f"Invalid component, expected str but found <{type(element).__name__}>: {element}"
        )

    return [element] + normversion(version)


def defaultversion(
    module: Optional[str], version: VersionType, element: "Element"
) -> Tuple[Optional[str], VersionType]:
    if module is not None:
        return module, normversion(version)

    # handle in-session defaults
    if isinstance(element.default, (list, tuple)) and not isinstance(
        element.default[0], str
    ):
        return element.default[0], element.default[1:] + normversion(version)

    default_version = normversion(element.default)

    if len(default_version) == 0:
        return module, normversion(version)

    if isinstance(default_version[0], str) and not default_version[
        0
    ].startswith("~"):
        if module is None:
            return default_version[0], default_version[1:] + normversion(
                version
            )
        else:
            return module, default_version[1:] + normversion(version)

    return module, default_version + normversion(version)


def extract(
    compact_element: Union[str, List[Union[str, dict]], None]
) -> Tuple[Optional[str], Optional[List[Union[str, dict]]]]:
    if compact_element is None:
        return None, None

    if isinstance(compact_element, str):
        return compact_element, None

    if isinstance(compact_element, omegaconf.listconfig.ListConfig):
        compact_element = list(compact_element)

    if not isinstance(compact_element, (list, tuple)):
        raise ValueError(
            f"Invalid component defintion. Expected list or str but found {type(compact_element)}: {compact_element}"
        )

    if len(compact_element) == 0:
        raise ValueError(
            f"Invalid component defintion. Expected str or non-empty list."
        )

    if not isinstance(compact_element[0], str):
        raise ValueError(
            f"Invalid component defintion. First element in list has to be a string."
        )

    if len(compact_element) == 1:
        return compact_element[0], None

    return compact_element[0], normversion(compact_element[1:])


def extend(
    module: Union[str, "Element", collections.abc.Iterable, None] = None,
    version: VersionType = None,
):
    if module is None or isinstance(module, str):
        # default
        return module, version

    if inspect.isclass(module):
        # in-session
        return module, version

    if isinstance(module, Element):
        # object instance
        return module.module, module.version() + normversion(version)

    # iterable
    m, v = extract(module)
    return m, normversion(v) + normversion(version)


def equaljson(a: Any, b: Any) -> bool:
    return json.dumps(a, sort_keys=True, default=serialize) == json.dumps(
        b, sort_keys=True, default=serialize
    )


def equalversion(a: VersionType, b: VersionType) -> bool:
    return equaljson(normversion(a), normversion(b))


def transfer_to(src: "Element", destination: "Element") -> "Element":
    destination.__model__ = src.__model__

    return destination


def uuid_to_id(uuid: str) -> str:
    alphabet = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"

    # convert uuid hex to base 62 of length 6
    result = ""
    for i in range(0, 6 * 2, 2):
        result += alphabet[int(uuid[i : i + 2], 16) % 62]

    return result


def instantiate(
    module: str, class_: "Element", version: VersionType, **constructor_kwargs
):
    try:
        Element._module_ = module  # assign project-relative module
        instance = class_(version=version, **constructor_kwargs)
        instance.on_instantiate()
        return instance
    except TypeError as _ex:
        raise MachinableError(
            f"Could not instantiate element {class_.__module__}.{class_.__name__}"
        ) from _ex


_CONNECTIONS = collections.defaultdict(list)


class Element(Mixin, Jsonable):
    """Element baseclass"""

    kind: Optional[str] = "Element"
    default: Optional["Element"] = None
    _module_: Optional[str] = None

    def __init__(self, version: VersionType = None):
        super().__init__()
        if Element._module_ is None:
            Element._module_ = self.__module__

        # resolve callables
        if not isinstance(
            version,
            (
                list,
                tuple,
                omegaconf.listconfig.ListConfig,
            ),
        ):
            version = [version]
        version = [v(self) if callable(v) else v for v in version]

        self.__model__ = getattr(schema, self.kind, schema.Element)(
            kind=self.kind,
            module=Element._module_,
            version=normversion(version),
            lineage=get_lineage(self),
            _dump=pickle.dumps(self.__class__)
            if Element._module_.startswith("__session__")
            else None,
        )
        self.__mixin__ = None
        self.__mixins__ = {}
        self._config: Optional[DictConfig] = None
        self._predicate: Optional[DictConfig] = None
        self._context: Optional[DictConfig] = None
        self._attributes = {}
        self._cache = {}
        self._kwargs = {}

        Element._module_ = None

    @property
    def uuid(self) -> str:
        return self.__model__.uuid

    @property
    def id(self) -> str:
        return id_from_uuid(self.uuid)

    @property
    def hash(self) -> str:
        return self.__model__.hash

    @property
    def timestamp(self) -> int:
        return self.__model__.timestamp

    def created_at(self) -> DatetimeType:
        return arrow.get(self.__model__.timestamp / 1e9)

    def version(
        self, version: VersionType = sentinel, overwrite: bool = False
    ) -> List[Union[str, dict]]:
        if version is sentinel:
            return self.__model__.version

        if hasattr(self, "is_mounted") and self.is_mounted():
            raise MachinableError(
                f"Cannot change version of mounted element {self}"
            )

        if overwrite:
            self.__model__.version = normversion(version)
        else:
            self.__model__.version.extend(normversion(version))

        self._clear_caches()

        return self.__model__.version

    @classmethod
    def set_default(
        cls,
        module: Union[str, "Element", None] = None,
        version: VersionType = None,
    ) -> None:
        if module is not None and not isinstance(module, str):
            cls.default = [module] + normversion(version)
            return
        cls.default = compact(module, version)

    def as_default(self) -> Self:
        cls = getattr(machinable, self.kind, Element)
        cls.set_default(self.__model__.module, self.__model__.version)

        return self

    @classmethod
    def connected(cls) -> List["Element"]:
        return _CONNECTIONS[cls.kind]

    @classmethod
    def get(
        cls,
        module: Union[str, "Element", None] = None,
        version: VersionType = None,
        **kwargs,
    ) -> "Element":
        if module is None and version is None:
            if len(_CONNECTIONS[cls.kind]) > 0:
                return _CONNECTIONS[cls.kind][-1]

        if module is None:
            return cls.make(module, version, **kwargs)

        return cls.singleton(module, version, **kwargs)

    @classmethod
    def make(
        cls,
        module: Union[None, str, "Element"] = None,
        version: VersionType = None,
        base_class: "Element" = None,
        **kwargs,
    ) -> "Element":
        if module is None:
            module = cls.__module__

        if base_class is None:
            base_class = getattr(machinable, cls.kind, Element)

        # prevent circular instantiation
        if module == "machinable" or module == base_class.__module__:
            return instantiate(module, base_class, version, **kwargs)

        from machinable.project import Project

        return Project.get().element(
            module=module,
            version=version,
            base_class=base_class,
            **kwargs,
        )

    @classmethod
    def instance(
        cls,
        module: Union[None, str, "Element"] = None,
        version: VersionType = None,
        **kwargs,
    ) -> "Element":
        module, version = defaultversion(module, version, cls)
        return cls.make(module, version, base_class=cls, **kwargs)

    @classmethod
    def singleton(
        cls,
        module: Union[str, "Element"],
        version: VersionType = None,
        **kwargs,
    ) -> "Element":
        # no-op as elements do not have a storage representation
        return cls.make(module, version, **kwargs)

    @classmethod
    def from_model(cls, model: schema.Element) -> "Element":
        if cls.__module__ != model.module:
            # re-instantiate element class
            if model.module and model.module.startswith("__session__"):
                if model._dump is None:
                    raise RuntimeError(
                        f"Unable to restore element {model.module}"
                    )
                instance = cls.make(pickle.loads(model._dump))
            else:
                instance = cls.make(model.module)
        else:
            instance = cls()

        instance.set_model(model)

        return instance

    def compute_context(self) -> Optional[Dict]:
        """Computes the context contraints of the element

        Returns:
            Optional[Dict]: Maps constraint fields their required JSON-able values.
                An empty dict means no constraits, i.e. all elements will be matched.
                Returning None means full constraits, i.e. no element will be matched.
        """
        return {
            "module": self.module,
            "config": {
                k: to_dict(v)
                for k, v in self.config.items()
                if k not in ["_default_", "_version_", "_update_"]
            },
            "predicate": self.compute_predicate(),
        }

    def on_compute_predicate(self) -> Dict:
        """Event to compute additional predicates that identify this element.

        Returns:
            Dict: Maps the names used during lookup to their JSON-able value.
        """
        return {}

    def compute_predicate(self) -> Dict:
        predicate = self.on_compute_predicate() or {}

        # apply scopes
        for scope in _CONNECTIONS["Scope"]:
            predicate.update(scope())

        return predicate

    @property
    def predicate(self) -> DictConfig:
        if self._predicate is None:
            if not self.__model__.predicate:
                return OmegaConf.create({})
            self._predicate = OmegaConf.create(self.__model__.predicate)

        return self._predicate

    @property
    def context(self) -> Optional[DictConfig]:
        if self._context is None:
            if self.__model__.context is None:
                return None
            self._context = OmegaConf.create(self.__model__.context)

        return self._context

    @property
    def config(self) -> DictConfig:
        """Element configuration"""
        if self._config is None:
            try:
                if self.__model__.config is not None:
                    self._config = OmegaConf.create(self.__model__.config)
                else:
                    # config method resolver
                    OmegaConf.register_new_resolver(
                        name="config_method",
                        resolver=ConfigMethod(self),
                        replace=True,
                    )

                    # expose default config so events and config methods can use it
                    default_config, config_model = from_element(self)

                    self._config = OmegaConf.create(default_config)

                    self.on_before_configure(self._config)

                    # apply versioning
                    __version = copy.deepcopy(self.__model__.version)

                    # compose configuration update
                    config_update = {}
                    for version in __version:
                        if isinstance(version, str) and version.startswith("~"):
                            definition = version[1:]

                            if not definition.endswith(")"):
                                definition = definition + "()"

                            method = match_method(definition)

                            if method is None:
                                raise ConfigurationError(
                                    f"Invalid version: {definition}"
                                )

                            version = ConfigMethod(self, "version")(*method)
                            if version is None:
                                version = {}

                            if not isinstance(version, collections.abc.Mapping):
                                raise ConfigurationError(
                                    f"Version method {definition} must produce a mapping, but returned {type(version)}: {version}"
                                )

                        config_update = update_dict(config_update, version)

                    # apply update
                    self._config = OmegaConf.merge(self._config, config_update)

                    # enable config methods
                    self._config = OmegaConf.create(
                        rewrite_config_methods(self._config)
                    )

                    # computed configuration transform
                    self.on_configure()

                    # resolve config
                    config = OmegaConf.to_container(self._config, resolve=True)

                    # enforce schema
                    if config_model is not None:
                        config = config_model(**config).model_dump()

                    # add introspection data
                    config["_default_"] = default_config
                    config["_version_"] = __version
                    config["_update_"] = config_update

                    # make config property accessible for slot components
                    self._config = OmegaConf.create(config)

                    # disallow further transformation
                    OmegaConf.set_readonly(self._config, True)

                    # save to model
                    self.__model__.config = OmegaConf.to_container(self._config)

                    OmegaConf.clear_resolver("config_method")

                    self.on_after_configure()
            except Exception as _ex:
                raise ConfigurationError(str(_ex)) from _ex

        return self._config

    @property
    def module(self) -> Optional[str]:
        return self.__model__.module

    @property
    def lineage(self) -> Tuple[str, ...]:
        return self.__model__.lineage

    @classmethod
    def collect(cls, elements) -> ElementCollection:
        return ElementCollection(elements)

    @classmethod
    def model(cls, element: Optional[Any] = None) -> schema.Element:
        if element is not None:
            if isinstance(element, cls):
                return element.__model__

            if isinstance(element, cls.model()):
                return element

            raise ValueError(f"Invalid {cls.kind.lower()} model: {element}.")

        if cls.kind is None:
            return schema.Element

        return getattr(schema, cls.kind)

    def matches(self, context: Optional[Dict] = None) -> bool:
        if context is None or self.context is None:
            # full constraint, match none
            return False

        if not context:
            # no constraints, match all
            return True

        for field, value in context.items():
            if field == "predicate":
                if hasattr(value, "items") and len(value) > 0:
                    for p, v in value.items():
                        if field not in self.context:
                            return False
                        if p not in self.context[field]:
                            return False
                        if not equaljson(self.context[field][p], v):
                            return False
            else:
                if field not in self.context:
                    return False
                if not equaljson(self.context[field], value):
                    return False

        return True

    def set_model(self, model: schema.Element) -> Self:
        self.__model__ = model

        # invalidate cached config
        self._config = None
        self._predicate = None

        return self

    def serialize(self) -> Dict:
        # ensure that configuration has been parsed and predicated computed
        assert self.config is not None
        self.__model__.predicate = self.compute_predicate()
        return self.__model__.model_dump()

    def save_attribute(self, name: Union[str, List[str]], data: Any) -> str:
        name = joinpath(name)

        self._attributes[name] = data

        return name

    def load_attribute(
        self, name: Union[str, List[str]], default=None
    ) -> Optional[Any]:
        name = joinpath(name)

        data = self._attributes.get(name, None)

        return data if data is not None else default

    @classmethod
    def unserialize(cls, serialized):
        return cls.from_model(cls.model()(**serialized))

    @classmethod
    def is_connected(cls) -> bool:
        return len(_CONNECTIONS[cls.kind]) > 0

    def on_instantiate(self) -> None:
        """Event that is invoked whenever the element is instantiated"""

    def on_before_configure(self, config: DictConfig) -> None:
        """Configuration event operating on the raw configuration
        This may be used to apply computed configuration updates
        Do not use to validate the configuration but use validators in the config schema
        that are applied at a later stage.
        """

    def on_configure(self) -> None:
        """Configuration event
        This may be used to apply computed configuration updates
        Do not use to validate the configuration but use validators in the config schema
        that are applied at a later stage.
        """

    def on_after_configure(self) -> None:
        """Configuration event operating on the resolved, read-only configuration
        This may be used to validate given the interface instance. If validation
        can be performed without the instance, use a config schema instead.
        """

    def _clear_caches(self) -> None:
        self._config = None
        self.__model__.config = None

    def __getattr__(self, name) -> Any:
        attr = getattr(self.__mixin__, name, None)
        if attr is not None:
            return attr
        return self.__getattribute__(name)

    def __enter__(self) -> Self:
        _CONNECTIONS[self.kind].append(self)
        return self

    def __exit__(self, *args, **kwargs):
        try:
            _CONNECTIONS[self.kind].pop()
        except IndexError:
            pass

    def __reduce__(self) -> Union[str, Tuple[Any, ...]]:
        return (self.__class__, (), self.serialize())

    def __getstate__(self):
        return self.serialize()

    def __setstate__(self, state):
        self.__model__ = self.__class__.model()(**state)

    def __repr__(self):
        return (
            f"{self.kind if self.module is None else self.module} [{self.id}]"
        )

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        return self.uuid == getattr(other, "uuid", None)

    def __ne__(self, other):
        return self.uuid != getattr(other, "uuid", None)


def get_lineage(element: "Element") -> Tuple[str, ...]:
    return tuple(
        obj.module if isinstance(obj, Element) else obj.__module__
        for obj in element.__class__.__mro__[1:-3]
    )


def get_dump(element: "Element") -> Optional[bytes]:
    if element.__model__.module.startswith("__session__"):
        return pickle.dumps(element.__class__)


def reset_connections() -> None:
    _CONNECTIONS.clear()



from typing import Any, Callable, Dict, List, Literal, Optional, Union

import copy
import json
import os
import shutil
import time
from contextlib import contextmanager
from functools import partial

try:
    from pysqlite3 import dbapi2 as sqlite3
except ModuleNotFoundError:
    import sqlite3

from machinable import schema
from machinable.element import get_lineage
from machinable.interface import Interface
from machinable.types import VersionType
from machinable.utils import is_directory_version, load_file, normjson


def interface_row_factory(row) -> schema.Interface:
    model = getattr(schema, row[1], schema.Interface)
    _version = json.loads(row[6])
    _config = json.loads(row[3])
    if _config is not None:
        _config["_update_"] = json.loads(row[4])
        _config["_default_"] = json.loads(row[5])
        _config["_version_"] = _version
    return model(
        uuid=row[0],
        kind=row[1],
        module=row[2],
        config=_config,
        version=_version,
        predicate=json.loads(row[7]),
        context=json.loads(row[11] or "null"),
        lineage=json.loads(row[8]),
        timestamp=int(row[9]),
        **json.loads(row[10]),
    )


def migrate(db: sqlite3.Connection) -> None:
    cur = db.cursor()
    version = cur.execute("PRAGMA user_version;").fetchone()[0]
    if version == 0:
        # initial migration
        cur.execute(
            """CREATE TABLE 'index' (
                uuid text PRIMARY KEY,
                kind text,
                module text,
                config json,
                config_update json,
                config_default json,
                version json,
                predicate json,
                lineage json,
                'timestamp' bigint,
                extra json
            )"""
        )
        cur.execute(
            """CREATE TABLE 'relations' (
                id integer PRIMARY KEY NOT NULL,
                relation text NOT NULL,
                uuid text NOT NULL REFERENCES 'index' (uuid),
                related_uuid text NOT NULL REFERENCES 'index' (uuid),
                'priority' integer NOT NULL DEFAULT 0,
                'timestamp' bigint,
                UNIQUE (uuid, related_uuid, relation)
            )"""
        )
        cur.execute("PRAGMA user_version = 1;")
        db.commit()
        version += 1
    if version == 1:
        # updates
        cur.execute("""ALTER TABLE 'index' ADD COLUMN 'context' json""")
        cur.execute("PRAGMA user_version = 2;")
        db.commit()
        version += 1
    if version == 2:
        # future migrations
        ...

    cur.close()


def load(database: str, create=False) -> sqlite3.Connection:
    if database.startswith("sqlite:///"):
        database = database[10:]
    if ":memory:" not in database:
        database = os.path.expanduser(database)
        if not os.path.isfile(database):
            if not create:
                raise FileNotFoundError(
                    f"Could not find index database at {database}"
                )
            else:
                os.makedirs(os.path.dirname(database), exist_ok=True)
    db = sqlite3.connect(database)
    migrate(db)
    return db


@contextmanager
def db(database: str, create=False) -> Optional[sqlite3.Connection]:
    try:
        database = load(database, create)
        yield database
        database.close()
    except FileNotFoundError:
        yield None


class Index(Interface):
    kind = "Index"
    default = None

    class Config:
        directory: str = "./storage"
        database: str = "in_directory('index.sqlite')"

    def __init__(self, version: VersionType = None):
        if is_directory_version(version):
            # interpret as shortcut for directory
            version = {"directory": version}
        super().__init__(version=version)
        self.__model__ = schema.Index(
            kind=self.kind,
            module=self.__model__.module,
            config=self.__model__.config,
            version=self.__model__.version,
            lineage=get_lineage(self),
        )

    def config_in_directory(self, relative_path: str) -> str:
        return os.path.join("sqlite:///", self.config.directory, relative_path)

    def local_directory(self, uuid: str, *append: str) -> str:
        # TODO: allow custom directory mappings
        return os.path.join(self.config.directory, uuid, *append)

    def commit(self, model: schema.Interface) -> bool:
        with db(self.config.database, create=True) as _db:
            cur = _db.cursor()
            if cur.execute(
                """SELECT uuid FROM 'index' WHERE uuid=?""", (model.uuid,)
            ).fetchone():
                cur.close()
                # already exists
                return False
            config = copy.deepcopy(model.config)
            if config:
                default = config.pop("_default_", {})
                version = config.pop("_version_", [])
                update = config.pop("_update_", {})
            else:
                default = {}
                version = []
                update = {}

            cur.execute(
                """INSERT INTO 'index' (
                    uuid,
                    kind,
                    module,
                    config,
                    config_update,
                    config_default,
                    version,
                    predicate,
                    context,
                    lineage,
                    'timestamp',
                    extra
                ) VALUES (?,?,?,?,?,?,?,?,?,?,?,?)""",
                (
                    model.uuid,
                    model.kind,
                    model.module,
                    normjson(config),
                    normjson(update),
                    normjson(default),
                    normjson(version),
                    normjson(model.predicate),
                    normjson(model.context),
                    normjson(model.lineage),
                    model.timestamp,
                    normjson(model.extra()),
                ),
            )
            _db.commit()
            cur.close()
        return True

    def create_relation(
        self,
        relation: str,
        uuid: str,
        related_uuid: Union[str, List[str]],
        priority: int = 0,
        timestamp: Optional[int] = None,
    ) -> None:
        if isinstance(related_uuid, (list, tuple)):
            for r in related_uuid:
                self.create_relation(relation, uuid, r, priority, timestamp)
            return
        with db(self.config.database, create=True) as _db:
            cur = _db.cursor()
            if cur.execute(
                """SELECT id FROM 'relations' WHERE uuid=? AND related_uuid=? AND relation=?""",
                (uuid, related_uuid, relation),
            ).fetchone():
                cur.close()
                # already exists
                return
            if timestamp is None:
                timestamp = time.time_ns()
            cur.execute(
                """INSERT INTO 'relations' (
                    relation,
                    uuid,
                    related_uuid,
                    priority,
                    timestamp
                ) VALUES (?,?,?,?,?)""",
                (relation, uuid, related_uuid, priority, timestamp),
            )
            _db.commit()
            cur.close()

    def find(
        self,
        interface: Union[schema.Interface, Interface],
        by: Literal["id", "uuid", "hash"] = "hash",
    ) -> List[schema.Interface]:
        if by in ["id", "uuid"]:
            found = self.find_by_id(interface.uuid)
            if not found:
                return []
            return [found]

        if by == "hash":
            return self.find_by_hash(interface.hash)

        raise ValueError(
            f"Invalid field: {by}. Must be one of 'id', 'uuid', 'hash'."
        )

    def find_by_id(self, uuid: str) -> Optional[schema.Interface]:
        with db(self.config.database, create=False) as _db:
            if not _db:
                return None
            cur = _db.cursor()
            if len(uuid) == 6:  # short id
                row = cur.execute(
                    """SELECT * FROM 'index' WHERE uuid LIKE ?""",
                    (f"%{uuid[:2]}-{uuid[2:]}-%",),
                ).fetchone()
            else:
                row = cur.execute(
                    """SELECT * FROM 'index' WHERE uuid=?""", (uuid,)
                ).fetchone()
            cur.close()
            if row is None:
                return None
            return interface_row_factory(row)

    def find_by_context(self, context: Dict) -> List[schema.Interface]:
        with db(self.config.database, create=False) as _db:
            if not _db:
                return []

            keys = []
            equals = []
            for field, value in context.items():
                if field == "predicate":
                    if hasattr(value, "items") and len(value) > 0:
                        # empty dict is a wildcard, so we only add
                        # condition if len(value) > 0
                        for p, v in value.items():
                            keys.append(
                                f"json_extract(context, '$.{field}.{p}')=?"
                            )
                            equals.append(v)
                else:
                    keys.append(f"json_extract(context, '$.{field}')=?")
                    equals.append(value)

            cur = _db.cursor()
            if len(keys) > 0:
                query = cur.execute(
                    """SELECT * FROM 'index' WHERE """ + (" AND ".join(keys)),
                    [
                        v if isinstance(v, (str, int, float)) else normjson(v)
                        for v in equals
                    ],
                ).fetchall()
            else:
                query = cur.execute("""SELECT * FROM 'index'""").fetchall()
            cur.close()
            return [interface_row_factory(row) for row in query]

    def find_by_hash(self, context_hash: str) -> List[schema.Interface]:
        with db(self.config.database, create=False) as _db:
            if not _db:
                return []

            cur = _db.cursor()
            query = cur.execute(
                """SELECT * FROM 'index' WHERE uuid LIKE ?""",
                (f"%{context_hash}",),
            ).fetchall()
            cur.close()

            return [interface_row_factory(row) for row in query]

    def find_related(
        self, relation: str, uuid: str, inverse: bool = False
    ) -> Union[None, List[schema.Interface]]:
        with db(self.config.database, create=False) as _db:
            if not _db:
                return None
            cur = _db.cursor()
            if not inverse:
                rows = cur.execute(
                    """SELECT * FROM 'index' WHERE uuid IN
                        (
                        SELECT related_uuid FROM relations WHERE uuid=? AND relation=?
                        )  ORDER BY 'timestamp' DESC
                    """,
                    (uuid, relation),
                ).fetchall()
            else:
                rows = cur.execute(
                    """SELECT * FROM 'index' WHERE uuid IN
                        (
                        SELECT uuid FROM relations WHERE related_uuid=? AND relation=?
                        )  ORDER BY 'timestamp' DESC
                    """,
                    (uuid, relation),
                ).fetchall()
            cur.close()
            return [interface_row_factory(row) for row in rows or []]

    def import_directory(
        self,
        directory: str,
        relations: bool = True,
        file_importer: Callable[[str, str], None] = partial(
            shutil.copytree, symlinks=True
        ),
    ):
        data = load_file([directory, "model.json"])
        model = getattr(schema, data["kind"])

        interface = model(**data)

        if relations:
            relations = load_file([directory, "related", "metadata.jsonl"], [])
        else:
            relations = []

        # transfer into local directory
        src = os.path.abspath(directory)
        dst = os.path.abspath(self.local_directory(interface.uuid))
        if src != dst:
            file_importer(src, dst)

        # create index entry
        self.commit(interface)
        for r in relations:
            self.create_relation(r["name"], r["uuid"], r["related_uuid"])



from machinable.collection import InterfaceCollection
from machinable.element import extend, normversion
from machinable.interface import Interface
from machinable.types import Optional, Union, VersionType


class Query:
    def by_id(self, uuid: str) -> Optional[Interface]:
        return Interface.find_by_id(uuid)

    def from_directory(self, directory: str) -> Interface:
        return Interface.from_directory(directory)

    def __call__(
        self,
        module: Union[str, Interface, None] = None,
        version: VersionType = None,
        **kwargs,
    ) -> Interface:
        module, version = extend(module, version)
        return Interface.get(module, version, **kwargs)

    # modifiers

    def all(
        self,
        module: Union[None, str, Interface] = None,
        version: VersionType = None,
        **kwargs,
    ) -> "InterfaceCollection":
        module, version = extend(module, version)
        return Interface.find(module, version, **kwargs)

    def new(
        self,
        module: Union[None, str, Interface] = None,
        version: VersionType = None,
        **kwargs,
    ) -> Interface:
        module, version = extend(module, version)
        return Interface.make(module, version, **kwargs)



from typing import Any, Dict, List, Optional, Union

import copy
import os
import sys
import time

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from typing import Literal

import arrow
from machinable import schema
from machinable.collection import ComponentCollection, ExecutionCollection
from machinable.component import Component
from machinable.element import extract, get_dump, get_lineage
from machinable.errors import ExecutionFailed
from machinable.interface import Interface, has_many, has_one
from machinable.schedule import Schedule
from machinable.types import (
    DatetimeType,
    ElementType,
    TimestampType,
    VersionType,
)
from machinable.utils import save_file, update_dict

_allowed_status = ["started", "heartbeat", "finished", "resumed"]


def _assert_allowed(status: str):
    if status not in _allowed_status:
        raise ValueError(
            f"Invalid status '{status}'; must be one of {_allowed_status}"
        )


StatusType = Literal["started", "heartbeat", "finished", "resumed"]


class Execution(Interface):
    kind = "Execution"
    default = None

    def __init__(
        self,
        version: VersionType = None,
        resources: Optional[Dict] = None,
        schedule: Union[Schedule, ElementType, None] = None,
    ):
        super().__init__(version)
        self.__model__ = schema.Execution(
            kind=self.kind,
            module=self.__model__.module,
            config=self.__model__.config,
            version=self.__model__.version,
            resources=resources,
            lineage=get_lineage(self),
        )
        self.__model__._dump = get_dump(self)
        if schedule is not None:
            if not isinstance(schedule, Schedule):
                schedule = Schedule.make(*extract(schedule))
            self.push_related("schedule", schedule)
        self._executable_ = None
        self._resources = {}
        self._defer_dispatch = False

    def deferred(self, defer: bool = True):
        self._defer_dispatch = defer
        return self

    @classmethod
    def collect(cls, executions) -> "ExecutionCollection":
        return ExecutionCollection(executions)

    def compute_context(self) -> Optional[Dict]:
        return None  # do not retrieve existing execution

    def on_compute_predicate(self) -> Dict:
        return {"resources": self.__model__.resources}

    @has_one
    def schedule() -> "Schedule":
        return Schedule

    @has_many(key="execution_history")
    def executables() -> ComponentCollection:
        return Component

    def executable(
        self, executable: Optional[Component] = None
    ) -> Optional[Component]:
        if executable is not None:
            return executable

        if self._executable_ is not None:
            return self._executable_

        # auto-select executable if only one is available
        if len(self.executables) == 1:
            return self.executables[0]

        raise ValueError(
            "No executable selected. Call `execution.of(executable)`, or pass an executable argument."
        )

    def of(self, executable: Union[None, Component]) -> Self:
        self._executable_ = executable
        return self

    @property
    def pending_executables(self) -> ComponentCollection:
        return self.executables.filter(lambda e: not e.cached())

    def on_add(self, executable: Component):
        """Event when executable is added"""

    def add(
        self,
        executable: Union[Component, List[Component]],
    ) -> Self:
        if isinstance(executable, (list, tuple)):
            for _executable in executable:
                self.add(_executable)
            return self

        if self.executables.contains(lambda x: x == executable):
            return self

        self.on_add(executable)

        self.push_related("executables", executable)

        return self

    def commit(self) -> Self:
        # ensure that configuration is parsed
        for executable in self.executables:
            assert executable.config is not None

        for executable in self.executables:
            executable.commit()

        return super().commit()

    def canonicalize_resources(self, resources: Dict) -> Dict:
        return resources

    def computed_resources(
        self, executable: Optional["Component"] = None
    ) -> Optional[Dict]:
        executable = self.executable(executable)

        if executable.id not in self._resources:
            # check if resources have been computed
            resources = self.load_file(
                [executable.id, "computed_resources.json"], default=None
            )
            if resources is None:
                # compute resources
                resources = self._compute_resources(executable)
                self.save_file(
                    [executable.id, "computed_resources.json"], resources
                )

            self._resources[executable.id] = resources

        return self._resources[executable.id]

    def _compute_resources(
        self, executable: Optional["Component"] = None
    ) -> Dict:
        default_resources = self.on_compute_default_resources(
            self.executable(executable)
        )

        if not self.__model__.resources and default_resources is not None:
            return self.canonicalize_resources(default_resources)

        if self.__model__.resources and not default_resources:
            resources = copy.deepcopy(self.__model__.resources)
            resources.pop("_inherit_defaults", None)
            return self.canonicalize_resources(resources)

        if self.__model__.resources and default_resources:
            resources = copy.deepcopy(self.__model__.resources)
            if resources.pop("_inherit_defaults", True) is False:
                return self.canonicalize_resources(resources)

            # merge with default resources
            defaults = self.canonicalize_resources(default_resources)
            update = self.canonicalize_resources(resources)

            defaults_ = copy.deepcopy(defaults)
            update_ = copy.deepcopy(update)

            # apply removals (e.g. #remove_me)
            removals = [
                k
                for k in update.keys()
                if isinstance(k, str) and k.startswith("#")
            ]
            for removal in removals:
                defaults_.pop(removal[1:], None)
                update_.pop(removal, None)

            return update_dict(defaults_, update_)

        return {}

    def dispatch(self) -> Self:
        if not self.executables:
            return self

        if len(self.pending_executables) == 0:
            return self

        if self.on_before_dispatch() is False:
            return self

        if self.on_verify_schedule() is False:
            raise ExecutionFailed(
                "The execution does not support the specified schedule."
            )

        self.commit()

        try:
            # compute resources
            self.pending_executables.map(lambda e: self.computed_resources(e))
            self.__call__()
            self.on_after_dispatch()
        except BaseException as _ex:  # pylint: disable=broad-except
            raise ExecutionFailed("Execution failed") from _ex

        return self

    def __call__(self) -> None:
        for executable in self.pending_executables:
            executable.dispatch()

    @property
    def host_info(
        self,
        executable: Optional["Component"] = None,
    ) -> Optional[Dict]:
        executable = self.executable(executable)
        return self.load_file([executable.id, "host.json"], None)

    def component_directory(self, *append: str, create: bool = False) -> str:
        return self.local_directory(
            self.executable().id, *append, create=create
        )

    def output_filepath(self, executable: Optional["Component"] = None) -> str:
        executable = self.executable(executable)
        return self.local_directory(executable.id, "output.log")

    def output(
        self,
        executable: Optional["Component"] = None,
        incremental: bool = False,
    ) -> Optional[str]:
        """Returns the output log"""
        if not self.is_mounted():
            return None

        executable = self.executable(executable)

        p = os.path.join(executable.id, "output.log")

        if incremental:
            read_length = self._cache.get(f"{p}.read_length", 0)
            if read_length == -1:
                return ""
            output = self.load_file(p, None)
            if output is None:
                return None

            if self.is_finished(executable):
                self._cache[f"{p}.read_length"] = -1
            else:
                self._cache[f"{p}.read_length"] = len(output)
            return output[read_length:]

        if p in self._cache:
            return self._cache[p]

        output = self.load_file(p, None)

        if self.is_finished(executable):
            self._cache[p] = output

        return output

    def stream_output(
        self,
        executable: Optional["Component"] = None,
        refresh_every: Union[int, float] = 1,
        stream=print,
    ):
        executable = self.executable(executable)
        try:
            while not self.is_started(executable) or self.is_active(executable):
                output = self.output(executable, incremental=True)
                if output:
                    stream(output)
                time.sleep(refresh_every)
        except KeyboardInterrupt:
            pass

    def update_status(
        self,
        executable: Optional["Component"] = None,
        status: StatusType = "heartbeat",
        timestamp: Optional[TimestampType] = None,
    ) -> TimestampType:
        _assert_allowed(status)
        executable = self.executable(executable)

        if timestamp is None:
            timestamp = arrow.now()
        if isinstance(timestamp, arrow.Arrow):
            timestamp = arrow.get(timestamp)

        # resumed event can occur multiple times
        multiple = status == "resumed"

        save_file(
            self.local_directory(executable.id, f"{status}_at"),
            str(timestamp) + ("\n" if multiple else ""),
            mode="a" if multiple else "w",
        )

        return timestamp

    def retrieve_status(
        self,
        executable: Optional["Component"] = None,
        status: StatusType = "heartbeat",
    ) -> Optional[DatetimeType]:
        _assert_allowed(status)
        executable = self.executable(executable)

        status = self.load_file([executable.id, f"{status}_at"], default=None)
        if status is None:
            return None

        multiple = status == "resumed"

        if multiple:
            # can have multiple rows, return latest
            status = status.strip("\n").split("\n")[-1]

        try:
            return arrow.get(status)
        except arrow.ParserError:
            return None

    def started_at(
        self, executable: Optional["Component"] = None
    ) -> Optional[DatetimeType]:
        """Returns the starting time"""
        if not self.is_mounted():
            return None
        return self.retrieve_status(self.executable(executable), "started")

    def resumed_at(
        self,
        executable: Optional["Component"] = None,
    ) -> Optional[DatetimeType]:
        """Returns the resumed time"""
        if not self.is_mounted():
            return None
        return self.retrieve_status(self.executable(executable), "resumed")

    def heartbeat_at(self, executable: Optional["Component"] = None):
        """Returns the last heartbeat time"""
        if not self.is_mounted():
            return None
        return self.retrieve_status(self.executable(executable), "heartbeat")

    def finished_at(
        self,
        executable: Optional["Component"] = None,
    ):
        """Returns the finishing time"""
        if not self.is_mounted():
            return None
        return self.retrieve_status(self.executable(executable), "finished")

    def is_finished(
        self,
        executable: Optional["Component"] = None,
    ):
        """True if finishing time has been written"""
        return bool(self.finished_at(executable))

    def is_started(
        self,
        executable: Optional["Component"] = None,
    ):
        """True if starting time has been written"""
        return bool(self.started_at(executable))

    def is_resumed(
        self,
        executable: Optional["Component"] = None,
    ):
        """True if resumed time has been written"""
        return bool(self.resumed_at(executable))

    def is_active(
        self,
        executable: Optional["Component"] = None,
    ):
        """True if not finished and last heartbeat occurred less than 30 seconds ago"""
        executable = self.executable(executable)

        if self.is_finished(executable):
            return False

        if not (
            heartbeat := self.heartbeat_at(
                executable,
            )
        ):
            return False

        return (arrow.now() - heartbeat).seconds < 30

    def is_live(
        self,
        executable: Optional["Component"] = None,
    ):
        """True if active or finished"""
        executable = self.executable(executable)
        return self.is_finished(executable) or self.is_active(executable)

    def is_incomplete(
        self,
        executable: Optional["Component"] = None,
    ):
        """Shorthand for is_started() and not is_live()"""
        executable = self.executable(executable)
        return self.is_started(executable) and not self.is_live(executable)

    def on_verify_schedule(self) -> bool:
        """Event to verify compatibility of the schedule"""
        if self.schedule is None:
            return True

        return False

    def on_before_dispatch(self) -> Optional[bool]:
        """Event triggered before dispatch of an execution

        Return False to prevent the dispatch
        """

    def on_after_dispatch(self) -> None:
        """Event triggered after the dispatch of an execution"""

    def on_compute_default_resources(
        self, executable: "Component"
    ) -> Optional[Dict]:
        """Event triggered to compute default resources"""

    @property
    def seed(self) -> int:
        return self.__model__.seed

    def __iter__(self):
        yield from self.executables

    def __exit__(self, *args):
        if len(args) == 3 and any(map(lambda x: x is not None, args)):
            # error occurred
            super().__exit__()
        elif self._defer_dispatch:
            super().__exit__()
        else:
            try:
                self.dispatch()
            finally:
                super().__exit__()
